import struct
from copy import copy
from inspect import signature
from typing import List, Any, NamedTuple, Callable, Dict, Union

from pykotor.common.geometry import Vector3
from pykotor.common.script import ScriptFunction, DataType
from pykotor.common.scriptdefs import KOTOR_FUNCTIONS
from pykotor.resource.formats.ncs import NCS, NCSInstruction, NCSInstructionType


class Interpreter:
    def __init__(self, ncs: NCS):
        self._ncs: NCS = ncs
        self._cursor: NCSInstruction = ncs.instructions[0]
        self._functions: List[ScriptFunction] = KOTOR_FUNCTIONS

        self._stack: Stack = Stack()
        self._returns: List[NCSInstruction] = [None]

        self._mocks: Dict[str, Callable] = {}

        self.stack_snapshots: List[StackSnapshot] = []
        self.action_snapshots: List[ActionSnapshot] = []

    def run(self):
        # TODO - limit how many instructions can be run before raising an error
        while self._cursor is not None:
            index = self._ncs.instructions.index(self._cursor)
            jump_value = None

            #print(str(index).ljust(3), str(self._cursor).ljust(40)[:40], str(self._stack.state()).ljust(30), f"BP={self._stack.base_pointer()//4}")

            if self._cursor.ins_type == NCSInstructionType.CONSTS:
                self._stack.add(DataType.STRING, self._cursor.args[0])

            elif self._cursor.ins_type == NCSInstructionType.CONSTI:
                self._stack.add(DataType.INT, self._cursor.args[0])

            elif self._cursor.ins_type == NCSInstructionType.CONSTF:
                self._stack.add(DataType.FLOAT, self._cursor.args[0])

            elif self._cursor.ins_type == NCSInstructionType.CONSTO:
                self._stack.add(DataType.OBJECT, self._cursor.args[0])

            elif self._cursor.ins_type == NCSInstructionType.CPTOPSP:
                self._stack.copy_to_top(self._cursor.args[0], self._cursor.args[1])

            elif self._cursor.ins_type == NCSInstructionType.CPDOWNSP:
                self._stack.copy_down(self._cursor.args[0], self._cursor.args[1])

            elif self._cursor.ins_type == NCSInstructionType.ACTION:
                self.do_action(self._functions[self._cursor.args[0]], self._cursor.args[1])

            elif self._cursor.ins_type == NCSInstructionType.MOVSP:
                self._stack.move(self._cursor.args[0])

            elif self._cursor.ins_type in [NCSInstructionType.ADDII, NCSInstructionType.ADDIF,
                                           NCSInstructionType.ADDFF, NCSInstructionType.ADDFI,
                                           NCSInstructionType.ADDSS, NCSInstructionType.ADDVV]:
                self._stack.addition_op()

            elif self._cursor.ins_type in [NCSInstructionType.SUBII, NCSInstructionType.SUBIF,
                                           NCSInstructionType.SUBFF, NCSInstructionType.SUBFI,
                                           NCSInstructionType.SUBVV]:
                self._stack.subtraction_op()

            elif self._cursor.ins_type in [NCSInstructionType.MULII, NCSInstructionType.MULIF,
                                           NCSInstructionType.MULFF, NCSInstructionType.MULFI,
                                           NCSInstructionType.MULVF, NCSInstructionType.MULFV]:
                self._stack.multiplication_op()

            elif self._cursor.ins_type in [NCSInstructionType.DIVII, NCSInstructionType.DIVIF,
                                           NCSInstructionType.DIVFF, NCSInstructionType.DIVFI,
                                           NCSInstructionType.DIVVF]:
                self._stack.division_op()

            elif self._cursor.ins_type in [NCSInstructionType.MODII]:
                self._stack.modulus_op()

            elif self._cursor.ins_type in [NCSInstructionType.NEGI, NCSInstructionType.NEGF]:
                self._stack.negation_op()

            elif self._cursor.ins_type in [NCSInstructionType.COMPI]:
                self._stack.bitwise_not_op()

            elif self._cursor.ins_type in [NCSInstructionType.NOTI]:
                self._stack.logical_not_op()

            elif self._cursor.ins_type in [NCSInstructionType.LOGANDII]:
                self._stack.logical_and_op()

            elif self._cursor.ins_type in [NCSInstructionType.LOGORII]:
                self._stack.logical_or_op()

            elif self._cursor.ins_type in [NCSInstructionType.INCORII]:
                self._stack.bitwise_or_op()

            elif self._cursor.ins_type in [NCSInstructionType.EXCORII]:
                self._stack.bitwise_xor_op()

            elif self._cursor.ins_type in [NCSInstructionType.BOOLANDII]:
                self._stack.bitwise_and_op()

            elif self._cursor.ins_type in [NCSInstructionType.EQUALII, NCSInstructionType.EQUALFF,
                                           NCSInstructionType.EQUALSS, NCSInstructionType.EQUALOO]:
                self._stack.logical_equality_op()

            elif self._cursor.ins_type in [NCSInstructionType.NEQUALII, NCSInstructionType.NEQUALFF,
                                           NCSInstructionType.NEQUALSS, NCSInstructionType.NEQUALOO]:
                self._stack.logical_inequality_op()

            elif self._cursor.ins_type in [NCSInstructionType.GTII, NCSInstructionType.GTFF]:
                self._stack.compare_greaterthan_op()

            elif self._cursor.ins_type in [NCSInstructionType.GEQII, NCSInstructionType.GEQFF]:
                self._stack.compare_greaterthanorequal_op()

            elif self._cursor.ins_type in [NCSInstructionType.LTII, NCSInstructionType.LTFF]:
                self._stack.compare_lessthan_op()

            elif self._cursor.ins_type in [NCSInstructionType.LEQII, NCSInstructionType.LEQFF]:
                self._stack.compare_lessthanorequal_op()

            elif self._cursor.ins_type in [NCSInstructionType.SHLEFTII]:
                self._stack.bitwise_leftshift_op()

            elif self._cursor.ins_type in [NCSInstructionType.SHRIGHTII]:
                self._stack.bitwise_rightshift_op()

            elif self._cursor.ins_type in [NCSInstructionType.INCIBP]:
                self._stack.increment_bp(self._cursor.args[0])

            elif self._cursor.ins_type in [NCSInstructionType.DECIBP]:
                self._stack.decrement_bp(self._cursor.args[0])

            elif self._cursor.ins_type in [NCSInstructionType.INCISP]:
                self._stack.increment(self._cursor.args[0])

            elif self._cursor.ins_type in [NCSInstructionType.DECISP]:
                self._stack.decrement(self._cursor.args[0])

            elif self._cursor.ins_type in [NCSInstructionType.RSADDI]:
                self._stack.add(DataType.INT, 0)

            elif self._cursor.ins_type in [NCSInstructionType.RSADDF]:
                self._stack.add(DataType.FLOAT, 0)

            elif self._cursor.ins_type in [NCSInstructionType.RSADDS]:
                self._stack.add(DataType.STRING, "")

            elif self._cursor.ins_type in [NCSInstructionType.RSADDO]:
                self._stack.add(DataType.OBJECT, 1)

            elif self._cursor.ins_type in [NCSInstructionType.RSADDEFF]:
                self._stack.add(DataType.EFFECT, 0)

            elif self._cursor.ins_type in [NCSInstructionType.RSADDTAL]:
                self._stack.add(DataType.TALENT, 0)

            elif self._cursor.ins_type in [NCSInstructionType.RSADDLOC]:
                self._stack.add(DataType.LOCATION, 0)

            elif self._cursor.ins_type in [NCSInstructionType.RSADDEVT]:
                self._stack.add(DataType.EVENT, 0)

            elif self._cursor.ins_type in [NCSInstructionType.SAVEBP]:
                self._stack.save_bp()

            elif self._cursor.ins_type in [NCSInstructionType.RESTOREBP]:
                self._stack.restore_bp()

            elif self._cursor.ins_type in [NCSInstructionType.CPTOPBP]:
                self._stack.copy_top_bp(self._cursor.args[0], self._cursor.args[1])

            elif self._cursor.ins_type in [NCSInstructionType.CPDOWNBP]:
                self._stack.copy_down_bp(self._cursor.args[0], self._cursor.args[1])

            elif self._cursor.ins_type in [NCSInstructionType.JSR]:
                index_return_to = self._ncs.instructions.index(self._cursor) + 1
                return_to = self._ncs.instructions[index_return_to]
                self._returns.append(return_to)

            elif self._cursor.ins_type in [NCSInstructionType.JZ, NCSInstructionType.JNZ]:
                jump_value = self._stack.pop()

            elif self._cursor.ins_type in [NCSInstructionType.STORE_STATE]:
                self.store_state()

            self.stack_snapshots.append(StackSnapshot(self._cursor, self._stack.state()))

            # Control flow
            if self._cursor.ins_type in [NCSInstructionType.RETN]:
                return_to = self._returns.pop()
                self._cursor = return_to
                continue

            elif self._cursor.ins_type in [NCSInstructionType.JMP]:
                self._cursor = self._cursor.jump

            elif self._cursor.ins_type in [NCSInstructionType.JZ] and jump_value == 0:
                self._cursor = self._cursor.jump

            elif self._cursor.ins_type in [NCSInstructionType.JNZ] and jump_value != 0:
                self._cursor = self._cursor.jump

            elif self._cursor.ins_type in [NCSInstructionType.JSR]:
                self._cursor = self._cursor.jump

            else:
                self._cursor = self._ncs.instructions[index + 1]

    def store_state(self):
        self._stack.store_state()

        index = self._ncs.instructions.index(self._cursor)
        tempcursor = self._ncs.instructions[index+2]

        block = []
        while tempcursor.ins_type != NCSInstructionType.RETN:
            block.append(tempcursor)
            index = self._ncs.instructions.index(tempcursor)
            tempcursor = self._ncs.instructions[index + 1]

        self._stack.add(DataType.ACTION, ActionStackValue(block, self._stack.state()))

    def do_action(self, function: ScriptFunction, args: int):
        args_snap = []

        for i in range(args):
            if function.params[i].datatype == DataType.VECTOR:
                vector_object = StackObject(DataType.VECTOR, Vector3(self._stack.pop().value, self._stack.pop().value, self._stack.pop().value))
                args_snap.append(vector_object)
            else:
                args_snap.append(self._stack.pop())

        for i in range(args):
            if function.params[i].datatype != args_snap[i].data_type:
                raise ValueError(f"Invoked action '{function.name}' received the wrong data type for parameter '{function.params[i].name}' valued at '{args_snap[i]}'.")

        if function.returntype != DataType.VOID:
            if function.name in self._mocks:
                # Execute and return the value back from the mock
                value = self._mocks[function.name](*[arg.value for arg in args_snap])
            else:
                # Return value of None if no relevant mock is found
                value = None

            if function.returntype == DataType.VECTOR:
                if value is None:
                    self._stack.add(DataType.FLOAT, 0.0)
                    self._stack.add(DataType.FLOAT, 0.0)
                    self._stack.add(DataType.FLOAT, 0.0)
                else:
                    self._stack.add(DataType.FLOAT, value[0])
                    self._stack.add(DataType.FLOAT, value[1])
                    self._stack.add(DataType.FLOAT, value[2])
            else:
                self._stack.add(function.returntype, value)

        self.action_snapshots.append(ActionSnapshot(function.name, args_snap, None))

    def print(self):
        for snap in self.stack_snapshots:
            print(snap.instruction, "\n", snap.stack, "\n")

    def set_mock(self, function_name: str, mock: Callable):
        function = next((function for function in self._functions if function.name == function_name), None)

        if function is None:
            raise ValueError(f"Function '{function_name}' does not exist.")

        mock_param_count = len(signature(mock).parameters)
        routine_param_count = len(function.params)
        if mock_param_count != routine_param_count:
            raise ValueError(f"Function '{function_name}' expects {routine_param_count} parameters not {mock_param_count}.")

        self._mocks[function_name] = mock

    def remove_mock(self, function_name: str):
        self._mocks.pop(function_name)


class StackV2:
    def __init__(self):
        self._stack: bytearray = bytearray()
        self._base_pointer: int = 0
        self._base_pointer_saved: List[int] = []
        self._stack_types: List = []

    def state(self) -> bytearray:
        return copy(self._stack)

    def copy_down(self, offset: int, size: int) -> None:
        stacksize = len(self._stack)
        copied = self._stack[stacksize-size:stacksize]
        self._stack[stacksize-offset:stacksize-offset+size] = copied

    def copy_to_top(self, offset: int, size: int) -> None:
        stacksize = len(self._stack)
        copied = self._stack[stacksize-offset:stacksize-offset+size]
        self._stack.extend(copied)

    def add(self, datatype: DataType, value: Union[int, float]):
        if datatype == DataType.INT:
            if not isinstance(value, int):
                raise ValueError
            self._stack.extend(struct.pack("i", value))
        elif datatype == DataType.FLOAT:
            if not isinstance(value, int):
                raise ValueError
            self._stack.extend(struct.pack("i", value))
        else:
            raise NotImplementedError


class Stack:
    def __init__(self):
        self._stack: List[StackObject] = []
        self._bp: int = 0
        self._bp_buffer: List[int] = []

    def state(self) -> List:
        return copy(self._stack)

    def add(self, data_type: DataType, value: Any) -> None:
        self._stack.append(StackObject(data_type, value))

    def _stack_index(self, offset: int) -> int:
        if offset >= 0:
            raise ValueError
        offset = abs(offset)
        index = 0
        while offset > 0:
            element_size = self._stack[index].data_type.size()
            offset -= element_size
            index -= 1
        return index

    def _stack_index_bp(self, offset) -> int:
        if offset >= 0:
            raise ValueError
        bp_index = self._bp // 4
        relative_index = abs(offset) // 4
        absolute_index = bp_index - relative_index
        if absolute_index < 0:
            raise ValueError
        return bp_index - relative_index

    def stack_pointer(self) -> int:
        return len(self._stack) * 4

    def base_pointer(self) -> int:
        return self._bp

    def peek(self, offset: int) -> Any:
        real_index = self._stack_index(offset)
        return self._stack[real_index]

    def copy_to_top(self, offset: int, size: int) -> None:
        for i in range(size//4):
            self._stack.append(self.peek(offset))

    def copy_down(self, offset: int, size: int) -> None:
        top_value = self._stack[-1]
        self._stack[self._stack_index(offset)] = top_value

    def pop(self) -> Any:
        return self._stack.pop()

    def move(self, offset: int) -> None:
        if offset > 0:
            raise ValueError
        if offset == 0:
            return
        remove_to = self._stack_index(offset)
        self._stack = self._stack[:remove_to]

    def copy_down_bp(self, offset: int, size: int) -> None:
        # Copy from the top of the stack down to the bp adjusted w/ offset?
        top_value = self._stack[-1]
        to_index = self._stack_index_bp(offset)
        self._stack[to_index] = top_value

    def copy_top_bp(self, offset: int, size: int) -> None:
        # Copy value relative to base pointer to the top of the stack
        copy_index = self._stack_index_bp(offset)
        top_value = self._stack[copy_index]
        self._stack.append(top_value)

    def save_bp(self) -> None:
        self._bp_buffer.append(self.base_pointer())
        self._bp = self.stack_pointer()

    def restore_bp(self) -> None:
        self._bp = self._bp_buffer.pop()

    def increment(self, offset: int):
        index = self._stack_index(offset)
        self._stack[index] = copy(self._stack[index])
        self._stack[index].value += 1

    def decrement(self, offset: int):
        index = self._stack_index(offset)
        self._stack[index] = copy(self._stack[index])
        self._stack[index].value -= 1

    def increment_bp(self, offset: int):
        index = self._stack_index_bp(offset)
        self._stack[index] = copy(self._stack[index])
        self._stack[index].value += 1

    def decrement_bp(self, offset: int):
        index = self._stack_index_bp(offset)
        self._stack[index] = copy(self._stack[index])
        self._stack[index].value -= 1

    def addition_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value2.data_type, value2.value + value1.value)

    def subtraction_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value2.data_type, value2.value - value1.value)

    def multiplication_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value2.value * value1.value)

    def division_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value2.value / value1.value)

    def modulus_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value2.value % value1.value)

    def negation_op(self):
        value1 = self._stack.pop()
        self.add(value1.data_type, -value1.value)

    def logical_not_op(self):
        value1 = self._stack.pop()
        self.add(value1.data_type, not value1.value)

    def logical_and_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value1.value and value2.value)

    def logical_or_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value1.value or value2.value)

    def logical_equality_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value1.value == value2.value)

    def logical_inequality_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value1.value != value2.value)

    def bitwise_not_op(self):
        value1 = self._stack.pop()
        self.add(value1.data_type, ~value1.value)

    def bitwise_or_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value1.value | value2.value)

    def bitwise_xor_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value1.value ^ value2.value)

    def bitwise_and_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value1.value & value2.value)

    def bitwise_leftshift_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value2.value << value1.value)

    def bitwise_rightshift_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, value2.value >> value1.value)

    def compare_greaterthan_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, int(value2.value > value1.value))

    def compare_greaterthanorequal_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, int(value2.value >= value1.value))

    def compare_lessthan_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, int(value2.value < value1.value))

    def compare_lessthanorequal_op(self):
        value1 = self._stack.pop()
        value2 = self._stack.pop()
        self.add(value1.data_type, int(value2.value <= value1.value))

    def store_state(self):
        ...


class StackObject:
    def __init__(self, data_type: DataType, value: Any):
        self.data_type: DataType = data_type
        self.value = value

    def __repr__(self):
        return f"{self.data_type.name}={self.value}"

    def __eq__(self, other):
        if isinstance(other, StackObject):
            return self.value == other.value
        else:
            return self.value == other


class ActionStackValue(NamedTuple):
    block: List[NCSInstruction]
    stack: List


class ActionSnapshot(NamedTuple):
    function_name: str
    arg_values: List
    return_value: Any


class StackSnapshot(NamedTuple):
    instruction: NCSInstruction
    stack: List[StackObject]


class EngineRoutineMock:
    def __init__(self, function: ScriptFunction, mock: Callable):
        self.function: ScriptFunction = function
        self.mock: Callable = mock



from __future__ import annotations

from typing import Optional

from pykotor.resource.formats.lip import LIP, LIPShape
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class LIPBinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._lip: Optional[LIP] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> LIP:
        self._lip = LIP()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "LIP ":
            raise TypeError("The file type that was loaded is invalid.")

        if file_version != "V1.0":
            raise TypeError("The LIP version that was loaded is not supported.")

        self._lip.length = self._reader.read_single()
        entry_count = self._reader.read_uint32()

        for i in range(entry_count):
            time = self._reader.read_single()
            shape = LIPShape(self._reader.read_uint8())
            self._lip.add(time, shape)

        return self._lip


class LIPBinaryWriter(ResourceWriter):
    HEADER_SIZE = 16
    LIP_ENTRY_SIZE = 5

    def __init__(
            self,
            lip: LIP,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._lip: LIP = lip

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        self._writer.write_string("LIP ")
        self._writer.write_string("V1.0")
        self._writer.write_single(self._lip.length)
        self._writer.write_uint32(len(self._lip))

        for keyframe in self._lip:
            self._writer.write_single(keyframe.time)
            self._writer.write_uint8(keyframe.shape.value)



import os.path
from typing import List, Optional, Dict

from pykotor.common.stream import BinaryReader
from pykotor.extract.file import FileResource, ResourceResult, ResourceIdentifier
from pykotor.resource.formats.erf import ERF, read_erf, write_erf, ERFType
from pykotor.resource.formats.rim import read_rim, write_rim, RIM
from pykotor.resource.type import ResourceType
from pykotor.tools.misc import is_rim_file, is_erf_file, is_capsule_file


class Capsule:
    """
    Chitin object is used for loading the list of resources stored in the .erf/.rim/.mod files used by the game.
    Resource data is not actually stored in memory by default but is instead loaded up on demand with the
    Capsule.resource() method.
    """
    def __init__(
            self,
            path: str,
            create_nonexisting: bool = False
    ):
        self._path: str = path
        self._resources: List[FileResource] = []

        if not is_capsule_file(path):
            raise ValueError(f"Invalid file extension in capsule filepath '{path}'.")

        if create_nonexisting and not os.path.exists(path):
            if is_rim_file(path):
                write_rim(RIM(), path)
            elif is_erf_file(path):
                write_erf(ERF(ERFType.from_extension(path)), path)

        self.reload()

    def __iter__(
            self
    ):
        for resource in self._resources:
            yield resource

    def __len__(
            self
    ):
        return len(self._resources)

    def resource(
            self,
            resref: str,
            restype: ResourceType,
            reload: bool = False
    ) -> Optional[bytes]:
        """
        Returns the bytes data of the specified resource. If the resource does not exist then returns None instead.

        Args:
            resref: The resource ResRef.
            restype: The resource type.
            reload: If True Capsule will reload the ERF before opening rather than using cached offsets.

        Returns:
            None or bytes data of resource.
        """
        if reload:
            self.reload()

        query = ResourceIdentifier(resref, restype)
        resource = next((resource for resource in self._resources if resource == query), None)
        return None if resource is None else resource.data()

    def batch(
            self,
            queries: List[ResourceIdentifier],
            reload: bool = False
    ) -> Dict[ResourceIdentifier, Optional[ResourceResult]]:
        if reload:
            self.reload()

        results = {}
        reader = BinaryReader.from_file(self.path())

        for query in queries:
            results[query] = None
            if self.exists(query.resname, query.restype):
                resource = next((resource for resource in self._resources if resource == query), None)
                if resource is not None:
                    reader.seek(resource.offset())
                    data = reader.read_bytes(resource.size())
                    results[query] = ResourceResult(query.resname, query.restype, self.path(), data)

        reader.close()
        return results

    def exists(
            self,
            resref: str,
            restype: ResourceType,
            reload: bool = False
    ) -> bool:
        if reload:
            self.reload()

        query = ResourceIdentifier(resref, restype)
        resource = next((resource for resource in self._resources if resource == query), None)
        return resource is not None

    def info(
            self,
            resref: str,
            restype: ResourceType,
            reload: bool = False
    ) -> FileResource:
        if reload:
            self.reload()

        query = ResourceIdentifier(resref, restype)
        resource = next((resource for resource in self._resources if resource == query), None)
        return resource

    def reload(
            self
    ):
        """
        Reload the list of resource info linked from the module file.
        """
        with BinaryReader.from_file(self._path) as reader:
            file_type = reader.read_string(4)
            file_version = reader.read_string(4)

            if file_type in ("ERF ", "MOD "):
                self._load_erf(reader)
            elif file_type == "RIM ":
                self._load_rim(reader)
            else:
                raise ValueError("File '{}' was not an ERF/MOD/RIM.".format(self._path))

    def add(
            self,
            resname: str,
            restype: ResourceType,
            resdata: bytes
    ):
        container = read_rim(self._path) if self._path.endswith(".rim") else read_erf(self._path)
        container.set(resname, restype, resdata)
        write_rim(container, self._path) if self._path.endswith(".rim") else write_erf(container, self._path)

    def path(
            self
    ) -> str:
        return self._path

    def _load_erf(
            self,
            reader: BinaryReader
    ):
        reader.skip(8)
        entry_count = reader.read_uint32()
        reader.skip(4)
        offset_to_keys = reader.read_uint32()
        offset_to_resources = reader.read_uint32()

        resrefs = []
        resids = []
        restypes = []
        reader.seek(offset_to_keys)
        for i in range(entry_count):
            resrefs.append(reader.read_string(16))
            resids.append(reader.read_uint32())
            restypes.append(ResourceType.from_id(reader.read_uint16()))
            reader.skip(2)

        reader.seek(offset_to_resources)
        for i in range(entry_count):
            res_offset = reader.read_uint32()
            res_size = reader.read_uint32()
            self._resources.append(FileResource(resrefs[i], restypes[i], res_size, res_offset, self._path))

    def _load_rim(
            self,
            reader: BinaryReader
    ):
        reader.skip(4)
        entry_count = reader.read_uint32()
        offset_to_entries = reader.read_uint32()

        reader.seek(offset_to_entries)
        for i in range(entry_count):
            resref = reader.read_string(16)
            restype = ResourceType.from_id(reader.read_uint32())
            res_id = reader.read_uint32()
            offset = reader.read_uint32()
            size = reader.read_uint32()
            self._resources.append(FileResource(resref, restype, size, offset, self._path))




from abc import ABC, abstractmethod
from typing import List, Union

from pykotor.resource.formats.ssf import SSF, SSFSound
from pykotor.tslpatcher.memory import PatcherMemory, TokenUsage


class ModifySSF:
    def __init__(self, sound: SSFSound, stringref: TokenUsage):
        self.sound: SSFSound = sound
        self.stringref: TokenUsage = stringref

    def apply(self, ssf: SSF, memory: PatcherMemory) -> None:
        ssf.set(self.sound, int(self.stringref.value(memory)))


class ModificationsSSF:
    def __init__(self, filename: str, replace_file: bool, modifiers: List[ModifySSF] = None):
        self.filename: str = filename
        self.replace_file: bool = replace_file
        self.modifiers: List[ModifySSF] = modifiers if modifiers is not None else []

    def apply(self, ssf: SSF, memory: PatcherMemory) -> None:
        for modifier in self.modifiers:
            modifier.apply(ssf, memory)




from typing import Optional

from pykotor.extract.installation import Installation
from pykotor.resource.formats.twoda import TwoDA
from pykotor.resource.generics.utp import UTP
from pykotor.resource.type import ResourceType


def get_model(
        utp: UTP,
        installation: Installation,
        *,
        placeables: Optional[TwoDA] = None,
) -> str:
    """
    Returns the model name for the given placeable.

    If no value is specified for the placeable parameters then it will be loaded from the given installation.

    Args:
        utp: UTC object of the target placeable.
        installation: The relevant installation.
        placeables: The placeables.2da loaded into a TwoDA object.

    Returns:
        Returns the model name for the placeable.
    """
    if placeables is None:
        placeables = installation.resource("genericdoors", ResourceType.TwoDA)

    model = placeables.get_row(utp.appearance_id).get_string("modelname")

    return model



"""
This module holds classes relating to string localization.
"""
from __future__ import annotations

from enum import IntEnum
from typing import Dict, Tuple, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    pass


class Language(IntEnum):
    """
    Language IDs recognized by both the games.
    """
    ENGLISH = 0
    FRENCH = 1
    GERMAN = 2
    ITALIAN = 3
    SPANISH = 4
    POLISH = 5

    @classmethod
    def _missing_(cls, value: int):
        return Language.ENGLISH


class Gender(IntEnum):
    """
    Gender IDs recognized by both the games in regards to string localization.
    """
    MALE = 0  # or neutral
    FEMALE = 1


class LocalizedString:
    """
    Localized strings are a way of the game handling strings that need to be catered to a specific language or gender.
    This is achieved through either referencing a entry in the 'dialog.tlk' or by directly providing strings for each
    language.

    Attributes:
        stringref: An index into the 'dialog.tlk' file. If this value is -1 the game will use the stored substrings.
    """

    def __init__(
            self,
            stringref: int
    ):
        self.stringref: int = stringref
        self._substrings: Dict[int] = {}

    def __iter__(
            self
    ):
        """
        Iterates through the list of substrings. Yields a tuple containing [language, gender, text]
        """
        for substring_id, text in self._substrings.items():
            yield *LocalizedString.substring_pair(substring_id), text

    def __len__(
            self
    ):
        """
        Returns the number of substrings.
        """
        return len(self._substrings)

    def __str__(
            self
    ):
        """
        If the stringref is valid, it will return it as a string. Otherwise it will return one of the substrings,
        prioritizing the english substring if it exists. If no substring exists and the stringref is invalid, "-1" is
        returned.
        """
        if self.stringref >= 0:
            return str(self.stringref)
        elif self.exists(Language.ENGLISH, Gender.MALE):
            return self.get(Language.ENGLISH, Gender.MALE)
        else:
            for language, gender, text in self:
                return text
            else:
                return "-1"

    def __eq__(
            self,
            other
    ):
        if not isinstance(other, LocalizedString):
            return False
        if other.stringref != self.stringref:
            return False
        if other._substrings != self._substrings:
            return False
        return True

    @classmethod
    def from_invalid(
            cls
    ) -> LocalizedString:
        return LocalizedString(-1)

    @classmethod
    def from_english(
            cls,
            text: str
    ) -> LocalizedString:
        """
        returns a new localizedstring object with a english substring.

        args:
            text: the text for the english substring.

        returns:
            a new localizedstring object.
        """
        locstring = LocalizedString(-1)
        locstring.set(Language.ENGLISH, Gender.MALE, text)
        return locstring

    @staticmethod
    def substring_id(
            language: Language,
            gender: Gender
    ) -> int:
        """
        Returns the ID for the language gender pair.

        Args:
            language: The language.
            gender: The gender.

        Returns:
            The substring ID.
        """
        return (language * 2) + gender

    @staticmethod
    def substring_pair(
            substring_id: int
    ) -> Tuple[Language, Gender]:
        """
        Returns the language gender pair from a substring ID.

        Args:
            substring_id: The substring ID.

        Returns:
            A tuple organized as (language, gender).
        """
        language = Language(substring_id // 2)
        gender = Gender(substring_id % 2)
        return language, gender

    def set(
            self,
            language: Language,
            gender: Gender,
            string: str
    ) -> None:
        """
        Sets the text of the substring with the corresponding language/gender pair. The substring is created if it does
        not exist.

        Args:
            language: The language.
            gender: The gender.
            string: The new text for the new substring.
        """
        substring_id = LocalizedString.substring_id(language, gender)
        self._substrings[substring_id] = string

    def get(
            self,
            language: Language,
            gender: Gender
    ) -> Optional[str]:
        """
        Gets the substring text with the corresponding language/gender pair.

        Args:
            language: The language.
            gender: The gender.

        Returns:
            The text of the substring if a matching pair is found, otherwise returns None.
        """
        substring_id = LocalizedString.substring_id(language, gender)
        return self._substrings[substring_id] if substring_id in self._substrings else None

    def remove(
            self,
            language: Language,
            gender: Gender
    ) -> None:
        """
        Removes the existing substring with the respective language/gender pair if it exists. No error is thrown if it
        does not find a corresponding pair.

        Args:
            language: The language.
            gender: The gender.
        """
        substring_id = LocalizedString.substring_id(language, gender)
        self._substrings.pop(substring_id)

    def exists(
            self,
            language: Language,
            gender: Gender
    ) -> bool:
        """
        Returns whether or not a substring exists with the respective language/gender pair.

        Args:
            language: The language.
            gender: The gender.

        Returns:
            True if the corresponding substring exists.
        """
        substring_id = LocalizedString.substring_id(language, gender)
        return substring_id in self._substrings



from __future__ import annotations

from typing import Optional, Iterator, List

from pykotor.common.geometry import Vector3, Vector4
from pykotor.resource.formats.lyt import LYT, LYTRoom, LYTTrack, LYTObstacle, LYTDoorHook
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class LYTAsciiReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._lyt: Optional[LYT] = None
        self._lines: List[str] = []

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> LYT:
        self._lyt = LYT()

        self._lines = self._reader.read_string(self._reader.size()).splitlines()

        iterator = iter(self._lines)
        for line in iterator:
            tokens = line.split()

            if tokens[0] == "roomcount":
                self._load_rooms(iterator, int(tokens[1]))
            if tokens[0] == "trackcount":
                self._load_tracks(iterator, int(tokens[1]))
            if tokens[0] == "obstaclecount":
                self._load_obstacles(iterator, int(tokens[1]))
            if tokens[0] == "doorhookcount":
                self._load_doorhooks(iterator, int(tokens[1]))

        if auto_close:
            self._reader.close()

        return self._lyt

    def _load_rooms(
            self,
            iterator: Iterator[str],
            count: int
    ):
        for i in range(count):
            tokens = next(iterator).split()
            model = tokens[0]
            position = Vector3(float(tokens[1]), float(tokens[2]), float(tokens[3]))
            self._lyt.rooms.append(LYTRoom(model, position))

    def _load_tracks(
            self,
            iterator: Iterator[str],
            count: int
    ):
        for i in range(count):
            tokens = next(iterator).split()
            model = tokens[0]
            position = Vector3(float(tokens[1]), float(tokens[2]), float(tokens[3]))
            self._lyt.tracks.append(LYTTrack(model, position))

    def _load_obstacles(
            self,
            iterator: Iterator[str],
            count: int
    ):
        for i in range(count):
            tokens = next(iterator).split()
            model = tokens[0]
            position = Vector3(float(tokens[1]), float(tokens[2]), float(tokens[3]))
            self._lyt.obstacles.append(LYTObstacle(model, position))

    def _load_doorhooks(
            self,
            iterator: Iterator[str],
            count: int
    ):
        for i in range(count):
            tokens = next(iterator).split()
            room = tokens[0]
            door = tokens[1]
            position = Vector3(float(tokens[3]), float(tokens[4]), float(tokens[5]))
            orientation = Vector4(float(tokens[6]), float(tokens[7]), float(tokens[8]), float(tokens[9]))
            self._lyt.doorhooks.append(LYTDoorHook(room, door, position, orientation))


class LYTAsciiWriter(ResourceWriter):
    def __init__(
            self,
            lyt: LYT,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._lyt: LYT = lyt

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        roomcount = len(self._lyt.rooms)
        trackcount = len(self._lyt.tracks)
        obstaclecount = len(self._lyt.obstacles)
        doorhookcount = len(self._lyt.doorhooks)

        self._writer.write_string("beginlayout\r\n")

        self._writer.write_string("   roomcount {}\r\n".format(roomcount))
        for room in self._lyt.rooms:
            self._writer.write_string("      {} {} {} {}\r\n".format(room.model,
                                                                     room.position.x, room.position.y, room.position.z))

        self._writer.write_string("   trackcount {}\r\n".format(trackcount))
        for track in self._lyt.tracks:
            self._writer.write_string("      {} {} {} {}\r\n".format(track.model,
                                                                     track.position.x, track.position.y,
                                                                     track.position.z))

        self._writer.write_string("   obstaclecount {}\r\n".format(obstaclecount))
        for obstacle in self._lyt.obstacles:
            self._writer.write_string("      {} {} {} {}\r\n".format(obstacle.model,
                                                                     obstacle.position.x, obstacle.position.y,
                                                                     obstacle.position.z))

        self._writer.write_string("   doorhookcount {}\r\n".format(doorhookcount))
        for doorhook in self._lyt.doorhooks:
            self._writer.write_string("      {} {} {} {} {} {} {} {} {} {}\r\n"
                                      .format(doorhook.room, doorhook.door, 0,
                                              doorhook.position.x, doorhook.position.y, doorhook.position.z,
                                              doorhook.orientation.x, doorhook.orientation.y, doorhook.orientation.z,
                                              doorhook.orientation.w))

        self._writer.write_string("donelayout")

        if auto_close:
            self._writer.close()



from pykotor.resource.formats.gff import GFF, GFFStruct, read_gff, GFFFieldType


class DiffGFF:
    def __init__(self, old: GFF, new: GFF):
        self.old: GFF = old
        self.new: GFF = new

    def is_same(self, old_struct: GFFStruct = None, new_struct: GFFStruct = None) -> bool:
        old_struct = self.old.root if old_struct is None else old_struct
        new_struct = self.new.root if new_struct is None else new_struct

        for label, ftype, old_value in old_struct:
            if not new_struct.exists(label):
                print("Missing a field:", label)
                return False

            if new_struct.what_type(label) != ftype:
                print("Field type has changed:", label)
                return False

            new_value = new_struct.value(label)
            if ftype == GFFFieldType.Struct:
                if old_value.struct_id != new_value.struct_id:
                    print("Struct ID has changed:", label)
                    return False

                if not self.is_same(old_value, new_value):
                    return False
            elif ftype == GFFFieldType.List:
                if len(old_value) != len(new_value):
                    print("List counts have changed:", label)
                    return False
                for i, old_child in enumerate(old_value):
                    new_child = new_value.at(i)
                    if not self.is_same(old_child, new_child):
                        return False
            else:
                if new_value != old_value:
                    print("Value has changed:", label, new_value, old_value)
                    return False

        return True



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.uts import construct_uts, dismantle_uts

TEST_FILE = "../../files/test.uts"


class TestUTS(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        uts = construct_uts(gff)
        self.validate_io(uts)

        gff = dismantle_uts(uts)
        uts = construct_uts(gff)
        self.validate_io(uts)

    def validate_io(self, uts):
        self.assertEqual("3Csounds", uts.tag)
        self.assertEqual(128551, uts.name.stringref)
        self.assertEqual("3csounds", uts.resref)
        self.assertEqual(1, uts.active)
        self.assertEqual(1, uts.continuous)
        self.assertEqual(1, uts.looping)
        self.assertEqual(1, uts.positional)
        self.assertEqual(1, uts.random_position)
        self.assertEqual(1, uts.random_position)
        self.assertEqual(1.5, uts.elevation)
        self.assertEqual(8.0, uts.max_distance)
        self.assertEqual(5.0, uts.min_distance)
        self.assertEqual(0.10000000149011612, uts.random_range_x)
        self.assertEqual(0.20000000298023224, uts.random_range_y)
        self.assertEqual(4000, uts.interval)
        self.assertEqual(100, uts.interval_variation)
        self.assertEqual(0.10000000149011612, uts.pitch_variation)
        self.assertEqual(22, uts.priority)
        self.assertEqual(0, uts.hours)
        self.assertEqual(3, uts.times)
        self.assertEqual(120, uts.volume)
        self.assertEqual(7, uts.volume_variation)
        self.assertEqual(6, uts.palette_id)
        self.assertEqual("comment", uts.comment)

        self.assertEqual(4, len(uts.sounds))
        self.assertEqual("c_drdastro_atk2", uts.sounds[3])




from __future__ import annotations

from typing import Optional

from pykotor.resource.formats.ssf.ssf_data import SSF, SSFSound
from pykotor.resource.type import TARGET_TYPES, SOURCE_TYPES, ResourceReader, ResourceWriter, autoclose


class SSFBinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._ssf: Optional[SSF] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> SSF:
        self._ssf = SSF()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "SSF ":
            raise ValueError("Attempted to load an invalid SSF was loaded.")

        if file_version != "V1.1":
            raise ValueError("The supplied SSF file version is not supported.")

        sounds_offset = self._reader.read_uint32()
        self._reader.seek(sounds_offset)

        self._ssf.set(SSFSound.BATTLE_CRY_1, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BATTLE_CRY_2, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BATTLE_CRY_3, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BATTLE_CRY_4, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BATTLE_CRY_5, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BATTLE_CRY_6, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.SELECT_1, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.SELECT_2, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.SELECT_3, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.ATTACK_GRUNT_1, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.ATTACK_GRUNT_2, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.ATTACK_GRUNT_3, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.PAIN_GRUNT_1, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.PAIN_GRUNT_2, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.LOW_HEALTH, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.DEAD, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.CRITICAL_HIT, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.TARGET_IMMUNE, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.LAY_MINE, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.DISARM_MINE, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BEGIN_STEALTH, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BEGIN_SEARCH, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.BEGIN_UNLOCK, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.UNLOCK_FAILED, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.UNLOCK_SUCCESS, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.SEPARATED_FROM_PARTY, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.REJOINED_PARTY, self._reader.read_uint32(max_neg1=True))
        self._ssf.set(SSFSound.POISONED, self._reader.read_uint32(max_neg1=True))

        return self._ssf


class SSFBinaryWriter(ResourceWriter):
    def __init__(
            self,
            ssf: SSF,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._ssf: SSF = ssf

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        self._writer.write_string("SSF ")
        self._writer.write_string("V1.1")
        self._writer.write_uint32(12)

        self._writer.write_uint32(self._ssf.get(SSFSound.BATTLE_CRY_1), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BATTLE_CRY_2), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BATTLE_CRY_3), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BATTLE_CRY_4), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BATTLE_CRY_5), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BATTLE_CRY_6), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.SELECT_1), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.SELECT_2), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.SELECT_3), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.ATTACK_GRUNT_1), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.ATTACK_GRUNT_2), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.ATTACK_GRUNT_3), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.PAIN_GRUNT_1), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.PAIN_GRUNT_2), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.LOW_HEALTH), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.DEAD), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.CRITICAL_HIT), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.TARGET_IMMUNE), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.LAY_MINE), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.DISARM_MINE), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BEGIN_STEALTH), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BEGIN_SEARCH), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.BEGIN_UNLOCK), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.UNLOCK_FAILED), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.UNLOCK_SUCCESS), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.SEPARATED_FROM_PARTY), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.REJOINED_PARTY), max_neg1=True)
        self._writer.write_uint32(self._ssf.get(SSFSound.POISONED), max_neg1=True)

        for i in range(12):
            self._writer.write_uint32(0xFFFFFFFF)



from unittest import TestCase

from pykotor.resource.formats.ncs import NCS
from pykotor.resource.formats.ncs.compiler.interpreter import Interpreter
from pykotor.resource.formats.ncs.compiler.lexer import NssLexer
from pykotor.resource.formats.ncs.compiler.parser import NssParser
from pykotor.resource.formats.ncs.optimizers import RemoveNopOptimizer


class TestNCSOptimizers(TestCase):

    def compile(self, script: str) -> NCS:
        nssLexer = NssLexer()
        nssParser = NssParser()
        lex = nssLexer.lexer
        parser = nssParser.parser

        t = parser.parse(script, tracking=True)

        ncs = NCS()
        t.compile(ncs)
        return ncs

    def test_no_op_optimizer(self):
        ncs = self.compile("""
            void main()
            {
                int value = 3;
                while (value > 0)
                {
                    if (value > 0)
                    {
                        PrintInteger(value);
                        value -= 1;
                    }
                }
            }
        """)

        ncs.optimize([RemoveNopOptimizer()])
        ncs.print()

        interpreter = Interpreter(ncs)
        interpreter.run()

        self.assertEqual(3, len(interpreter.action_snapshots))
        self.assertEqual(3, interpreter.action_snapshots[0].arg_values[0])
        self.assertEqual(2, interpreter.action_snapshots[1].arg_values[0])
        self.assertEqual(1, interpreter.action_snapshots[2].arg_values[0])



import contextlib
import errno
import logging
import logging.handlers
import os
import sys
from logging import Filter
from typing import IO, Any, Callable, Iterator, Optional, TextIO, Type, cast

from pip._internal.utils._log import VERBOSE, getLogger
from pip._internal.utils.compat import WINDOWS
from pip._internal.utils.deprecation import DEPRECATION_MSG_PREFIX
from pip._internal.utils.misc import ensure_dir

try:
    import threading
except ImportError:
    import dummy_threading as threading  # type: ignore


try:
    from pip._vendor import colorama
# Lots of different errors can come from this, including SystemError and
# ImportError.
except Exception:
    colorama = None


_log_state = threading.local()
subprocess_logger = getLogger("pip.subprocessor")


class BrokenStdoutLoggingError(Exception):
    """
    Raised if BrokenPipeError occurs for the stdout stream while logging.
    """


def _is_broken_pipe_error(exc_class: Type[BaseException], exc: BaseException) -> bool:
    if exc_class is BrokenPipeError:
        return True

    # On Windows, a broken pipe can show up as EINVAL rather than EPIPE:
    # https://bugs.python.org/issue19612
    # https://bugs.python.org/issue30418
    if not WINDOWS:
        return False

    return isinstance(exc, OSError) and exc.errno in (errno.EINVAL, errno.EPIPE)


@contextlib.contextmanager
def indent_log(num: int = 2) -> Iterator[None]:
    """
    A context manager which will cause the log output to be indented for any
    log messages emitted inside it.
    """
    # For thread-safety
    _log_state.indentation = get_indentation()
    _log_state.indentation += num
    try:
        yield
    finally:
        _log_state.indentation -= num


def get_indentation() -> int:
    return getattr(_log_state, "indentation", 0)


class IndentingFormatter(logging.Formatter):
    default_time_format = "%Y-%m-%dT%H:%M:%S"

    def __init__(
        self,
        *args: Any,
        add_timestamp: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        A logging.Formatter that obeys the indent_log() context manager.

        :param add_timestamp: A bool indicating output lines should be prefixed
            with their record's timestamp.
        """
        self.add_timestamp = add_timestamp
        super().__init__(*args, **kwargs)

    def get_message_start(self, formatted: str, levelno: int) -> str:
        """
        Return the start of the formatted log message (not counting the
        prefix to add to each line).
        """
        if levelno < logging.WARNING:
            return ""
        if formatted.startswith(DEPRECATION_MSG_PREFIX):
            # Then the message already has a prefix.  We don't want it to
            # look like "WARNING: DEPRECATION: ...."
            return ""
        if levelno < logging.ERROR:
            return "WARNING: "

        return "ERROR: "

    def format(self, record: logging.LogRecord) -> str:
        """
        Calls the standard formatter, but will indent all of the log message
        lines by our current indentation level.
        """
        formatted = super().format(record)
        message_start = self.get_message_start(formatted, record.levelno)
        formatted = message_start + formatted

        prefix = ""
        if self.add_timestamp:
            prefix = f"{self.formatTime(record)} "
        prefix += " " * get_indentation()
        formatted = "".join([prefix + line for line in formatted.splitlines(True)])
        return formatted


def _color_wrap(*colors: str) -> Callable[[str], str]:
    def wrapped(inp: str) -> str:
        return "".join(list(colors) + [inp, colorama.Style.RESET_ALL])

    return wrapped


class ColorizedStreamHandler(logging.StreamHandler):

    # Don't build up a list of colors if we don't have colorama
    if colorama:
        COLORS = [
            # This needs to be in order from highest logging level to lowest.
            (logging.ERROR, _color_wrap(colorama.Fore.RED)),
            (logging.WARNING, _color_wrap(colorama.Fore.YELLOW)),
        ]
    else:
        COLORS = []

    def __init__(self, stream: Optional[TextIO] = None, no_color: bool = None) -> None:
        super().__init__(stream)
        self._no_color = no_color

        if WINDOWS and colorama:
            self.stream = colorama.AnsiToWin32(self.stream)

    def _using_stdout(self) -> bool:
        """
        Return whether the handler is using sys.stdout.
        """
        if WINDOWS and colorama:
            # Then self.stream is an AnsiToWin32 object.
            stream = cast(colorama.AnsiToWin32, self.stream)
            return stream.wrapped is sys.stdout

        return self.stream is sys.stdout

    def should_color(self) -> bool:
        # Don't colorize things if we do not have colorama or if told not to
        if not colorama or self._no_color:
            return False

        real_stream = (
            self.stream
            if not isinstance(self.stream, colorama.AnsiToWin32)
            else self.stream.wrapped
        )

        # If the stream is a tty we should color it
        if hasattr(real_stream, "isatty") and real_stream.isatty():
            return True

        # If we have an ANSI term we should color it
        if os.environ.get("TERM") == "ANSI":
            return True

        # If anything else we should not color it
        return False

    def format(self, record: logging.LogRecord) -> str:
        msg = super().format(record)

        if self.should_color():
            for level, color in self.COLORS:
                if record.levelno >= level:
                    msg = color(msg)
                    break

        return msg

    # The logging module says handleError() can be customized.
    def handleError(self, record: logging.LogRecord) -> None:
        exc_class, exc = sys.exc_info()[:2]
        # If a broken pipe occurred while calling write() or flush() on the
        # stdout stream in logging's Handler.emit(), then raise our special
        # exception so we can handle it in main() instead of logging the
        # broken pipe error and continuing.
        if (
            exc_class
            and exc
            and self._using_stdout()
            and _is_broken_pipe_error(exc_class, exc)
        ):
            raise BrokenStdoutLoggingError()

        return super().handleError(record)


class BetterRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def _open(self) -> IO[Any]:
        ensure_dir(os.path.dirname(self.baseFilename))
        return super()._open()


class MaxLevelFilter(Filter):
    def __init__(self, level: int) -> None:
        self.level = level

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno < self.level


class ExcludeLoggerFilter(Filter):

    """
    A logging Filter that excludes records from a logger (or its children).
    """

    def filter(self, record: logging.LogRecord) -> bool:
        # The base Filter class allows only records from a logger (or its
        # children).
        return not super().filter(record)


def setup_logging(verbosity: int, no_color: bool, user_log_file: Optional[str]) -> int:
    """Configures and sets up all of the logging

    Returns the requested logging level, as its integer value.
    """

    # Determine the level to be logging at.
    if verbosity >= 2:
        level_number = logging.DEBUG
    elif verbosity == 1:
        level_number = VERBOSE
    elif verbosity == -1:
        level_number = logging.WARNING
    elif verbosity == -2:
        level_number = logging.ERROR
    elif verbosity <= -3:
        level_number = logging.CRITICAL
    else:
        level_number = logging.INFO

    level = logging.getLevelName(level_number)

    # The "root" logger should match the "console" level *unless* we also need
    # to log to a user log file.
    include_user_log = user_log_file is not None
    if include_user_log:
        additional_log_file = user_log_file
        root_level = "DEBUG"
    else:
        additional_log_file = "/dev/null"
        root_level = level

    # Disable any logging besides WARNING unless we have DEBUG level logging
    # enabled for vendored libraries.
    vendored_log_level = "WARNING" if level in ["INFO", "ERROR"] else "DEBUG"

    # Shorthands for clarity
    log_streams = {
        "stdout": "ext://sys.stdout",
        "stderr": "ext://sys.stderr",
    }
    handler_classes = {
        "stream": "pip._internal.utils.logging.ColorizedStreamHandler",
        "file": "pip._internal.utils.logging.BetterRotatingFileHandler",
    }
    handlers = ["console", "console_errors", "console_subprocess"] + (
        ["user_log"] if include_user_log else []
    )

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "filters": {
                "exclude_warnings": {
                    "()": "pip._internal.utils.logging.MaxLevelFilter",
                    "level": logging.WARNING,
                },
                "restrict_to_subprocess": {
                    "()": "logging.Filter",
                    "name": subprocess_logger.name,
                },
                "exclude_subprocess": {
                    "()": "pip._internal.utils.logging.ExcludeLoggerFilter",
                    "name": subprocess_logger.name,
                },
            },
            "formatters": {
                "indent": {
                    "()": IndentingFormatter,
                    "format": "%(message)s",
                },
                "indent_with_timestamp": {
                    "()": IndentingFormatter,
                    "format": "%(message)s",
                    "add_timestamp": True,
                },
            },
            "handlers": {
                "console": {
                    "level": level,
                    "class": handler_classes["stream"],
                    "no_color": no_color,
                    "stream": log_streams["stdout"],
                    "filters": ["exclude_subprocess", "exclude_warnings"],
                    "formatter": "indent",
                },
                "console_errors": {
                    "level": "WARNING",
                    "class": handler_classes["stream"],
                    "no_color": no_color,
                    "stream": log_streams["stderr"],
                    "filters": ["exclude_subprocess"],
                    "formatter": "indent",
                },
                # A handler responsible for logging to the console messages
                # from the "subprocessor" logger.
                "console_subprocess": {
                    "level": level,
                    "class": handler_classes["stream"],
                    "no_color": no_color,
                    "stream": log_streams["stderr"],
                    "filters": ["restrict_to_subprocess"],
                    "formatter": "indent",
                },
                "user_log": {
                    "level": "DEBUG",
                    "class": handler_classes["file"],
                    "filename": additional_log_file,
                    "encoding": "utf-8",
                    "delay": True,
                    "formatter": "indent_with_timestamp",
                },
            },
            "root": {
                "level": root_level,
                "handlers": handlers,
            },
            "loggers": {"pip._vendor": {"level": vendored_log_level}},
        }
    )

    return level_number



"""
This module holds various unrelated classes.
"""
from __future__ import annotations

import os
from enum import IntEnum, Enum
from typing import TypeVar, Generic, Dict

from pykotor.common.geometry import Vector3

T = TypeVar("T")


class ResRef:
    class InvalidEncodingError(ValueError):
        ...

    class ExceedsMaxLengthError(ValueError):
        ...

    """
    A string reference to a game resource. ResRefs can be a maximum of 16 characters in length.
    """

    def __init__(
            self,
            text: str
    ):
        self._value = ""
        self.set(text)

    def __len__(
            self
    ):
        return len(self._value)

    def __eq__(
            self,
            other
    ):
        """
        A ResRef can be compared to another ResRef or a str.
        """
        if isinstance(other, ResRef):
            return other.get().lower() == self.get().lower()
        elif isinstance(other, str):
            return other.lower() == self.get().lower()
        else:
            return NotImplemented

    def __repr__(
            self
    ):
        return "ResRef({})".format(self._value)

    def __str__(
            self
    ):
        return self._value

    @classmethod
    def from_blank(
            cls
    ) -> ResRef:
        """
        Returns a blank ResRef.

        Returns:
            A new ResRef instance.
        """
        return ResRef("")

    @classmethod
    def from_path(
            cls,
            path: str
    ) -> ResRef:
        """
        Returns a ResRef from the filename in the specified path.

        Args:
            path: The filepath.

        Returns:
            A new ResRef instance.
        """
        return ResRef(os.path.splitext(path)[0].replace('\\', '/').split('/')[-1])

    def set(
            self,
            text: str,
            truncate: bool = True
    ) -> None:
        """
        Sets the ResRef.

        Args:
            text: The reference string.
            truncate: If true, the string will be truncated to 16 characters, otherwise it will raise an error instead.

        Raises:
            ValueError:
        """
        if len(text) > 16 and truncate:
            text = text[:16]
        elif len(text) > 16 and not truncate:
            raise ResRef.ExceedsMaxLengthError("ResRef cannot exceed 16 characters.")
        if len(text) != len(text.encode()):
            raise ResRef.InvalidEncodingError("ResRef must be in ASCII characters.")

        self._value = text

    def get(
            self
    ) -> str:
        return self._value


class Game(IntEnum):
    """
    Represents which game.
    """
    K1 = 1
    K2 = 2


class Color:
    # Listed here for hinting purposes
    RED: Color
    GREEN: Color
    BLUE: Color
    BLACK: Color
    WHITE: Color

    def __init__(
            self,
            r: float,
            g: float,
            b: float,
            a: float = 1.0
    ):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    def __repr__(
            self
    ):
        return "Color({}, {}, {}, {})".format(self.r, self.g, self.b, self.g)

    def __str__(
            self
    ):
        """
        Returns a string of each color component separated by whitespace.
        """
        return "{} {} {} {}".format(self.r, self.g, self.b, self.a)

    def __eq__(
            self,
            other
    ):
        """
        Two Color instances are equal if their color components are equal.
        """
        if not isinstance(other, Color):
            return NotImplemented

        return other.r == self.r and other.g == self.g and other.b == self.b and other.a == self.a

    @classmethod
    def from_rgb_integer(
            cls,
            integer: int
    ) -> Color:
        """
        Returns a Color by decoding the specified integer.

        Args:
            integer: RGB integer.

        Returns:
            A new Color instance.
        """
        red = (0x000000FF & integer) / 255
        green = ((0x0000FF00 & integer) >> 8) / 255
        blue = ((0x00FF0000 & integer) >> 16) / 255
        return Color(red, green, blue)

    @classmethod
    def from_rgba_integer(
            cls,
            integer: int
    ) -> Color:
        """
        Returns a Color by decoding the specified integer.

        Args:
            integer: RGB integer.

        Returns:
            A new Color instance.
        """
        red = (0x000000FF & integer) / 255
        green = ((0x0000FF00 & integer) >> 8) / 255
        blue = ((0x00FF0000 & integer) >> 16) / 255
        alpha = ((0xFF000000 & integer) >> 24) / 255
        return Color(red, green, blue, alpha)

    @classmethod
    def from_bgr_integer(
            cls,
            integer: int
    ) -> Color:
        """
        Returns a Color by decoding the specified integer.

        Args:
            integer: BGR integer.

        Returns:
            A new Color instance.
        """
        red = ((0x00FF0000 & integer) >> 16) / 255
        green = ((0x0000FF00 & integer) >> 8) / 255
        blue = (0x000000FF & integer) / 255
        return Color(red, green, blue)

    @classmethod
    def from_rgb_vector3(
            cls,
            vector3: Vector3
    ) -> Color:
        """
        Returns a Color from the specified vector components.

        Args:
            vector3: A Vector3 instance.

        Returns:
            A new Color instance.
        """
        red = vector3.x
        green = vector3.y
        blue = vector3.z
        return Color(red, green, blue)

    @classmethod
    def from_bgr_vector3(
            cls,
            vector3: Vector3
    ) -> Color:
        """
        Returns a Color from the specified vector components.

        Args:
            vector3: A Vector3 instance.

        Returns:
            A new Color instance.
        """
        red = vector3.z
        green = vector3.y
        blue = vector3.x
        return Color(red, green, blue)

    def rgb_integer(
            self
    ) -> int:
        """
        Returns a RGB integer encoded from the color components.

        Returns:
            A integer representing a color.
        """
        red = int(self.r * 255) << 0
        green = int(self.g * 255) << 8
        blue = int(self.b * 255) << 16
        return red + green + blue

    def rgba_integer(
            self
    ) -> int:
        """
        Returns a RGB integer encoded from the color components.

        Returns:
            A integer representing a color.
        """
        red = int(self.r * 255) << 0
        green = int(self.g * 255) << 8
        blue = int(self.b * 255) << 16
        alpha = int(self.a * 255) << 24
        return red + green + blue + alpha

    def bgr_integer(
            self
    ) -> int:
        """
        Returns a BGR integer encoded from the color components.

        Returns:
            A integer representing a color.
        """
        red = int(self.r * 255) << 16
        green = int(self.g * 255) << 8
        blue = int(self.b * 255)
        return red + green + blue

    def rgb_vector3(
            self
    ) -> Vector3:
        """
        Returns a Vector3 representing a color with its components.

        Returns:
            A new Vector3 instance.
        """
        return Vector3(self.r, self.g, self.b)

    def bgr_vector3(
            self
    ) -> Vector3:
        """
        Returns a Vector3 representing a color with its components.

        Returns:
            A new Vector3 instance.
        """
        return Vector3(self.b, self.g, self.r)


Color.RED = Color(1.0, 0.0, 0.0)
Color.GREEN = Color(0.0, 1.0, 0.0)
Color.BLUE = Color(0.0, 0.0, 1.0)
Color.BLACK = Color(0.0, 0.0, 0.0)
Color.WHITE = Color(1.0, 1.0, 1.0)


class WrappedInt:
    def __init__(
            self,
            value: int = 0
    ):
        self._value: int = value

    def __add__(
            self,
            other
    ):
        if isinstance(other, WrappedInt):
            self._value += other.get()
        elif isinstance(other, int):
            self._value += other
        else:
            return NotImplemented

    def __eq__(
            self,
            other
    ):
        if isinstance(other, WrappedInt):
            return self.get() == other.get()
        elif isinstance(other, int):
            return self.get() == other
        else:
            return NotImplemented

    def set(
            self,
            value: int
    ) -> None:
        self._value = value

    def get(
            self
    ) -> int:
        return self._value


class InventoryItem:
    def __init__(
            self,
            resref: ResRef,
            droppable: bool = False,
            infinite: bool = False
    ):
        self.resref: ResRef = resref
        self.droppable: bool = droppable
        self.infinite: bool = infinite

    def __str__(
            self
    ):
        return self.resref.get()

    def __eq__(
            self,
            other: InventoryItem
    ):
        return self.resref == other.resref and self.droppable == other.droppable


class EquipmentSlot(Enum):
    INVALID = 0
    HEAD = 1 ** 0
    ARMOR = 2 ** 1
    GAUNTLET = 2 ** 3
    RIGHT_HAND = 2 ** 4
    LEFT_HAND = 2 ** 5
    RIGHT_ARM = 2 ** 7
    LEFT_ARM = 2 ** 8
    IMPLANT = 2 ** 9
    BELT = 2 ** 10
    CLAW1 = 2 ** 14
    CLAW2 = 2 ** 15
    CLAW3 = 2 ** 16
    HIDE = 2 ** 17
    # TSL Only:
    RIGHT_HAND_2 = 2 ** 18
    LEFT_HAND_2 = 2 ** 19


class CaseInsensitiveDict(Generic[T]):
    def __init__(
            self
    ):
        self._dictionary: Dict[str, T] = {}
        self._case_map: Dict[str, str] = {}

    def __getitem__(
            self,
            key: str
    ):
        return self._dictionary[key.lower()]

    def __setitem__(
            self,
            key: str,
            value: T
    ):
        self._dictionary[key.lower()] = value
        self._case_map[key.lower()] = key

    def __delitem__(
            self,
            key: str
    ):
        del self._dictionary[key.lower()]
        del self._case_map[key.lower()]

    def __contains__(
            self,
            key: str
    ):
        return key.lower() in self._dictionary

    def __len__(
            self
    ):
        return len(self._dictionary)

    def __repr__(
            self
    ):
        return repr(self._dictionary)

    def pop(
            self,
            key: str
    ):
        self._dictionary.pop(key.lower())
        self._case_map.pop(key.lower())

    def get(
            self,
            key: str
    ):
        return self._dictionary[key.lower()]

    def items(
            self
    ):
        return [(self._case_map[key], value) for key, value in self._dictionary.items()]

    def values(
            self
    ):
        return self._dictionary.values()

    def keys(
            self
    ):
        return [self._case_map[key] for key in self._dictionary.keys()]



from __future__ import annotations

from enum import IntEnum
from typing import List, Optional

from pykotor.common.geometry import Vector3
from pykotor.common.language import LocalizedString, Gender, Language
from pykotor.common.misc import Game, ResRef, Color
from pykotor.resource.formats.gff import GFF, GFFList, GFFStruct, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class DLG:
    """
    Stores dialog data.

    Attributes:
        word_count: "NumWords" field.
        on_abort: "EndConverAbort" field.
        on_end: "EndConversation" field.
        skippable: "Skippable" field.
        ambient_track: "AmbientTrack" field.
        animated_cut: "AnimatedCut" field.
        camera_model: "CameraModel" field.
        computer_type.value: "ComputerType" field.
        conversation_type.value: "ConversationType" field.
        old_hit_check: "OldHitCheck" field.
        unequip_hands: "UnequipHItem" field.
        unequip_items: "UnequipItems" field.
        vo_id: "VO_ID" field.

        alien_race_owner: "AlienRaceOwner" field. KotOR 2 Only.
        post_proc_owner: "PostProcOwner" field. KotOR 2 Only.
        record_no_vo: "RecordNoVO" field. KotOR 2 Only.
        next_node_id: "NextNodeID" field. KotOR 2 Only.

        delay_entry: "DelayEntry" field. Not used by the game engine.
        delay_reply: "DelayReply" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.DLG

    def __init__(
            self,
            blank_node: bool = True
    ):
        self.starters: List[DLGLink] = []
        self.stunts: List[DLGStunt] = []

        if blank_node:
            # Add bare minimum to be openable by DLGEditor
            starter = DLGLink()
            entry = DLGEntry()
            entry.text.set(Language.ENGLISH, Gender.MALE, "")
            starter.node = entry
            self.starters.append(starter)

        self.ambient_track: ResRef = ResRef.from_blank()
        self.animated_cut: int = 0
        self.camera_model: ResRef = ResRef.from_blank()
        self.computer_type: DLGComputerType = DLGComputerType.Modern
        self.conversation_type: DLGConversationType = DLGConversationType.Human
        self.on_abort: ResRef = ResRef.from_blank()
        self.on_end: ResRef = ResRef.from_blank()
        self.word_count: int = 0
        self.old_hit_check: bool = False
        self.skippable: bool = False
        self.unequip_items: bool = False
        self.unequip_hands: bool = False
        self.vo_id: str = ""

        # KotOR 2:
        self.alien_race_owner: int = 0
        self.next_node_id: int = 0
        self.post_proc_owner: int = 0
        self.record_no_vo: bool = False

        # Deprecated:
        self.delay_entry: int = 0
        self.delay_reply: int = 0

    def print_tree(
            self
    ) -> None:
        """
        Prints all the nodes (one per line) in the dialog tree with appropriate indentation.
        """
        self._print_tree(self.starters, 0, [], [])

    def _print_tree(
            self,
            links: List[DLGLink],
            indent: int,
            seen_links: List[DLGLink],
            seen_nodes: List[DLGNode]
    ):
        for link in links:
            if link.node not in seen_nodes:
                print("{}-> {}".format(" " * indent, link.node.text))
                seen_links.append(link)

                if link.node not in seen_nodes:
                    seen_nodes.append(link.node)
                    self._print_tree(link.node.links, indent + 3, seen_links, seen_nodes)
            else:
                print("{}-> [LINK] {}".format(" " * indent, link.node.text))

    def all_entries(
            self
    ) -> List[DLGEntry]:
        """
        Returns a flat list of all entries in the dialog.

        Returns:
            A list of all stored entries.
        """
        return self._all_entries()

    def _all_entries(
            self,
            links: List[DLGLink] = None,
            seen_entries: List = None
    ) -> List[DLGEntry]:
        entries = []

        links = self.starters if links is None else links
        seen_entries = [] if seen_entries is None else seen_entries

        for link in links:
            entry = link.node
            if entry not in seen_entries:
                entries.append(entry)
                seen_entries.append(entry)
                for reply_link in entry.links:
                    reply = reply_link.node
                    entries.extend(self._all_entries(reply.links, seen_entries))

        return entries

    def all_replies(
            self
    ) -> List[DLGReply]:
        """
        Returns a flat list of all replies in the dialog.

        Returns:
            A list of all stored replies.
        """
        return self._all_replies()

    def _all_replies(
            self,
            links: List[DLGLink] = None,
            seen_replies: List = None
    ) -> List[DLGReply]:
        replies = []

        links = [_ for link in self.starters for _ in link.node.links] if links is None else links
        seen_replies = [] if seen_replies is None else seen_replies

        for link in links:
            reply = link.node
            if reply not in seen_replies:
                replies.append(reply)
                seen_replies.append(reply)
                for entry_link in reply.links:
                    entry = entry_link.node
                    replies.extend(self._all_replies(entry.links, seen_replies))

        return replies


class DLGComputerType(IntEnum):
    Modern = 0
    Ancient = 1


class DLGConversationType(IntEnum):
    Human = 0
    Computer = 1
    Other = 2


class DLGNode:
    """
    Represents a node in the dialog tree.

    Attributes:
        text: "Text" field.
        listener: "Listener" field.
        vo_resref: "VO_ResRef" field.
        script1: "Script" field.
        delay: "Delay" field.
        comment: "Comment" field.
        sound: "Sound" field.
        quest: "Quest" field.
        plot_index: "PlotIndex" field.
        plot_xp_percentage: "PlotXPPercentage" field.
        wait_flags: "WaitFlags" field.
        camera_angle: "CameraAngle" field.
        fade_type: "FadeType" field.
        sound_exists: "SoundExists" field.
        vo_text_changed: "Changed" field.

        quest_entry: "QuestEntry" field.
        fade_delay: "FadeDelay" field.
        fade_length: "FadeLength" field.
        camera_anim: "CameraAnimation" field.
        camera_id: "CameraID" field.
        camera_fov: "CamFieldOfView" field.
        camera_height: "CamHeightOffset" field.
        camera_effect: "CamVidEffect" field.
        target_height: "TarHeightOffset" field.
        fade_color: "FadeColor" field.

        script1_param1: "ActionParam1" field. KotOR 2 Only.
        script2_param1: "ActionParam1b" field. KotOR 2 Only.
        script1_param2: "ActionParam2" field. KotOR 2 Only.
        script2_param2: "ActionParam2b" field. KotOR 2 Only.
        script1_param3: "ActionParam3" field. KotOR 2 Only.
        script2_param3: "ActionParam3b" field. KotOR 2 Only.
        script1_param4: "ActionParam4" field. KotOR 2 Only.
        script2_param4: "ActionParam4b" field. KotOR 2 Only.
        script1_param5: "ActionParam5" field. KotOR 2 Only.
        script2_param5: "ActionParam5b" field. KotOR 2 Only.
        script1_param6: "ActionParamStrA" field. KotOR 2 Only.
        script2_param6: "ActionParamStrB" field. KotOR 2 Only.
        script2: "Script2" field. KotOR 2 Only.
        alien_race_node: "AlienRaceNode" field. KotOR 2 Only.
        emotion_id: "Emotion" field. KotOR 2 Only.
        facial_id: "FacialAnim" field. KotOR 2 Only.
        node_id: "NodeID" field. KotOR 2 Only.
        unskippable: "NodeUnskippable" field. KotOR 2 Only.
        post_proc_node: "PostProcNode" field. KotOR 2 Only.
        record_no_vo_override: "RecordNoVOOverri" field. KotOR 2 Only.
        record_vo: "RecordVO" field. KotOR 2 Only.
        vo_text_changed: "VOTextChanged" field. KotOR 2 Only.
    """

    def __init__(
            self
    ):
        self.comment: str = ""
        self.links: List[DLGLink] = []

        self.camera_angle: int = 0
        self.delay: int = -1
        self.fade_type: int = 0
        self.listener: str = ""
        self.plot_index: int = 0
        self.plot_xp_percentage: float = 0.0
        self.quest: str = ""
        self.script1: ResRef = ResRef.from_blank()
        self.sound: ResRef = ResRef.from_blank()
        self.sound_exists: bool = False
        self.text: LocalizedString = LocalizedString.from_invalid()
        self.vo_resref: ResRef = ResRef.from_blank()
        self.wait_flags: int = 0

        self.animations: List[DLGAnimation] = []

        self.quest_entry: Optional[int] = 0
        self.fade_color: Optional[Color] = None
        self.fade_delay: Optional[float] = None
        self.fade_length: Optional[float] = None
        self.camera_anim: Optional[int] = None
        self.camera_id: Optional[int] = None
        self.camera_fov: Optional[float] = None
        self.camera_height: Optional[float] = None
        self.camera_effect: Optional[int] = None
        self.target_height: Optional[float] = None

        # KotOR 2:
        self.script1_param1: int = 0
        self.script1_param2: int = 0
        self.script1_param3: int = 0
        self.script1_param4: int = 0
        self.script1_param5: int = 0
        self.script1_param6: str = ""
        self.script2_param1: int = 0

        self.script2: ResRef = ResRef.from_blank()
        self.script2_param2: int = 0
        self.script2_param3: int = 0
        self.script2_param4: int = 0
        self.script2_param5: int = 0
        self.script2_param6: str = ""

        self.alien_race_node: int = 0
        self.emotion_id: int = 0
        self.facial_id: int = 0
        self.unskippable: bool = False
        self.node_id: int = 0
        self.post_proc_node: int = 0

        self.record_no_vo_override: bool = False
        self.record_vo: bool = False
        self.vo_text_changed: bool = False

    def __repr__(
            self
    ):
        return str(self.text.get(Language.ENGLISH, Gender.MALE))


class DLGReply(DLGNode):
    """
    Replies are nodes that are responses by the player.
    """

    def __init__(
            self
    ):
        super().__init__()


class DLGEntry(DLGNode):
    """
    Entries are nodes that are responses by NPCs.
    """

    def __init__(
            self
    ):
        super().__init__()
        self.speaker: str = ""


class DLGAnimation:
    """
    Represents a unit of animation executed during a node.
    """

    def __init__(
            self
    ):
        self.animation_id: int = 6
        self.participant: str = ""


class DLGLink:
    """
    Points to a node. Links are stored either in other nodes or in the starting list of the DLG.

    Attributes:
        active1: "Active" field.
        comment: "LinkComment" field. Only used in links stored in nodes.
        is_child: "IsChild" field. Only used in links stored in nodes.
        active2: "Active2" field. KotOR 2 Only.
        logic: "Logic" field. KotOR 2 Only.
        active1_not: "Not" field. KotOR 2 Only.
        active2_not: "Not2" field. KotOR 2 Only.
        active1_param1: "Param1" field. KotOR 2 Only.
        active1_param2: "Param2" field. KotOR 2 Only.
        active1_param3: "Param3" field. KotOR 2 Only.
        active1_param4: "Param4" field. KotOR 2 Only.
        active1_param5: "Param5" field. KotOR 2 Only.
        active1_param6: "ParamStrA" field. KotOR 2 Only.
        active2_param1: "Param1b" field. KotOR 2 Only.
        active2_param2: "Param2b" field. KotOR 2 Only.
        active2_param3: "Param3b" field. KotOR 2 Only.
        active2_param4: "Param4b" field. KotOR 2 Only.
        active2_param5: "Param5b" field. KotOR 2 Only.
        active2_param6: "ParamStrB" field. KotOR 2 Only.
    """

    def __init__(
            self,
            node: DLGNode = DLGNode
    ):
        self.active1: ResRef = ResRef.from_blank()
        self.node: DLGNode = node

        # not in StartingList
        self.is_child: bool = False
        self.comment: str = ""

        # KotOR 2 Only:
        self.active2: ResRef = ResRef.from_blank()
        self.active1_not: bool = False
        self.active2_not: bool = False
        self.logic: bool = False

        self.active1_param1: int = 0
        self.active1_param2: int = 0
        self.active1_param3: int = 0
        self.active1_param4: int = 0
        self.active1_param5: int = 0
        self.active1_param6: str = ""

        self.active2_param1: int = 0
        self.active2_param2: int = 0
        self.active2_param3: int = 0
        self.active2_param4: int = 0
        self.active2_param5: int = 0
        self.active2_param6: str = ""


class DLGStunt:
    """

    Attributes:
        participant: "Participant" field.
        stunt_model: "StuntModel" field.
    """

    def __init__(
            self
    ):
        self.participant: str = ""
        self.stunt_model: ResRef = ResRef.from_blank()


def construct_dlg(
        gff: GFF
) -> DLG:
    def construct_node(
            gff_struct: GFFStruct,
            node: DLGNode
    ):
        node.text = gff_struct.acquire("Text", LocalizedString.from_invalid())
        node.listener = gff_struct.acquire("Listener", "")
        node.vo_resref = gff_struct.acquire("VO_ResRef", ResRef.from_blank())
        node.script1 = gff_struct.acquire("Script", ResRef.from_blank())
        node.delay = -1 if gff_struct.acquire("Delay", 0) == 0xFFFFFFFF else gff_struct.acquire("Delay", 0)
        node.comment = gff_struct.acquire("Comment", "")
        node.sound = gff_struct.acquire("Sound", ResRef.from_blank())
        node.quest = gff_struct.acquire("Quest", "")
        node.plot_index = gff_struct.acquire("PlotIndex", -1)
        node.plot_xp_percentage = gff_struct.acquire("PlotXPPercentage", 0.0)
        node.wait_flags = gff_struct.acquire("WaitFlags", 0)
        node.camera_angle = gff_struct.acquire("CameraAngle", 0)
        node.fade_type = gff_struct.acquire("FadeType", 0)
        node.sound_exists = gff_struct.acquire("SoundExists", 0)
        node.vo_text_changed = gff_struct.acquire("Changed", 0)

        for anim_struct in gff_struct.acquire("AnimList", GFFList()):
            anim = DLGAnimation()
            anim.animation_id = anim_struct.acquire("Animation", 0)
            anim.participant = anim_struct.acquire("Participant", "")
            node.animations.append(anim)

        node.script1_param1 = gff_struct.acquire("ActionParam1", 0)
        node.script2_param1 = gff_struct.acquire("ActionParam1b", 0)
        node.script1_param2 = gff_struct.acquire("ActionParam2", 0)
        node.script2_param2 = gff_struct.acquire("ActionParam2b", 0)
        node.script1_param3 = gff_struct.acquire("ActionParam3", 0)
        node.script2_param3 = gff_struct.acquire("ActionParam3b", 0)
        node.script1_param4 = gff_struct.acquire("ActionParam4", 0)
        node.script2_param4 = gff_struct.acquire("ActionParam4b", 0)
        node.script1_param5 = gff_struct.acquire("ActionParam5", 0)
        node.script2_param5 = gff_struct.acquire("ActionParam5b", 0)
        node.script1_param6 = gff_struct.acquire("ActionParamStrA", "")
        node.script2_param6 = gff_struct.acquire("ActionParamStrB", "")
        node.script2 = gff_struct.acquire("Script2", ResRef.from_blank())
        node.alien_race_node = gff_struct.acquire("AlienRaceNode", 0)
        node.emotion_id = gff_struct.acquire("Emotion", 0)
        node.facial_id = gff_struct.acquire("FacialAnim", 0)
        node.node_id = gff_struct.acquire("NodeID", 0)
        node.unskippable = gff_struct.acquire("NodeUnskippable", 0)
        node.post_proc_node = gff_struct.acquire("PostProcNode", 0)
        node.record_no_vo_override = gff_struct.acquire("RecordNoVOOverri", 0)
        node.record_vo = gff_struct.acquire("RecordVO", 0)
        node.vo_text_changed = gff_struct.acquire("VOTextChanged", 0)

        if gff_struct.exists("QuestEntry"): node.quest_entry = gff_struct.acquire("QuestEntry", 0)
        if gff_struct.exists("FadeDelay"): node.fade_delay = gff_struct.acquire("FadeDelay", 0.0)
        if gff_struct.exists("FadeLength"): node.fade_length = gff_struct.acquire("FadeLength", 0.0)
        if gff_struct.exists("CameraAnimation"): node.camera_anim = gff_struct.acquire("CameraAnimation", 0)
        if gff_struct.exists("CameraID"): node.camera_id = gff_struct.acquire("CameraID", 0)
        if gff_struct.exists("CamFieldOfView"): node.camera_fov = gff_struct.acquire("CamFieldOfView", 0.0)
        if gff_struct.exists("CamHeightOffset"): node.camera_height = gff_struct.acquire("CamHeightOffset", 0.0)
        if gff_struct.exists("CamVidEffect"): node.camera_effect = gff_struct.acquire("CamVidEffect", 0)
        if gff_struct.exists("TarHeightOffset"): node.target_height = gff_struct.acquire("TarHeightOffset", 0.0)
        if gff_struct.exists("FadeColor"):
            node.fade_color = Color.from_bgr_vector3(gff_struct.acquire("FadeColor", Vector3.from_null()))

    def construct_link(
            gff_struct: GFFStruct,
            link: DLGLink
    ):
        link.active1 = gff_struct.acquire("Active", ResRef.from_blank())
        link.active2 = gff_struct.acquire("Active2", ResRef.from_blank())
        link.logic = gff_struct.acquire("Logic", 0)
        link.active1_not = gff_struct.acquire("Not", 0)
        link.active2_not = gff_struct.acquire("Not2", 0)
        link.active1_param1 = gff_struct.acquire("Param1", 0)
        link.active1_param2 = gff_struct.acquire("Param2", 0)
        link.active1_param3 = gff_struct.acquire("Param3", 0)
        link.active1_param4 = gff_struct.acquire("Param4", 0)
        link.active1_param5 = gff_struct.acquire("Param5", 0)
        link.active1_param6 = gff_struct.acquire("ParamStrA", "")
        link.active2_param1 = gff_struct.acquire("Param1b", 0)
        link.active2_param2 = gff_struct.acquire("Param2b", 0)
        link.active2_param3 = gff_struct.acquire("Param3b", 0)
        link.active2_param4 = gff_struct.acquire("Param4b", 0)
        link.active2_param5 = gff_struct.acquire("Param5b", 0)
        link.active2_param6 = gff_struct.acquire("ParamStrB", "")

    dlg = DLG(False)

    root = gff.root

    all_entries = [DLGEntry() for _ in range(len(root.acquire("EntryList", GFFList())))]
    all_replies = [DLGReply() for _ in range(len(root.acquire("ReplyList", GFFList())))]

    dlg.word_count = root.acquire("NumWords", 0)
    dlg.on_abort = root.acquire("EndConverAbort", ResRef.from_blank())
    dlg.on_end = root.acquire("EndConversation", ResRef.from_blank())
    dlg.skippable = bool(root.acquire("Skippable", 0))
    dlg.ambient_track = root.acquire("AmbientTrack", ResRef.from_blank())
    dlg.animated_cut = root.acquire("AnimatedCut", 0)
    dlg.camera_model = root.acquire("CameraModel", ResRef.from_blank())
    dlg.computer_type = DLGComputerType(root.acquire("ComputerType", 0))
    dlg.conversation_type = DLGConversationType(root.acquire("ConversationType", 0))
    dlg.old_hit_check = root.acquire("OldHitCheck", 0)
    dlg.unequip_hands = root.acquire("UnequipHItem", 0)
    dlg.unequip_items = root.acquire("UnequipItems", 0)
    dlg.vo_id = root.acquire("VO_ID", "")
    dlg.alien_race_owner = root.acquire("AlienRaceOwner", 0)
    dlg.post_proc_owner = root.acquire("PostProcOwner", 0)
    dlg.record_no_vo = root.acquire("RecordNoVO", 0)
    dlg.next_node_id = root.acquire("NextNodeID", 0)
    dlg.delay_entry = root.acquire("DelayEntry", 0)
    dlg.delay_reply = root.acquire("DelayReply", 0)

    for stunt_struct in root.acquire("StuntList", GFFList()):
        stunt = DLGStunt()
        dlg.stunts.append(stunt)
        stunt.participant = stunt_struct.acquire("Participant", "")
        stunt.stunt_model = stunt_struct.acquire("StuntModel", ResRef.from_blank())

    for link_struct in root.acquire("StartingList", GFFList()):
        entry_index = link_struct.acquire("Index", 0)
        link = DLGLink()
        dlg.starters.append(link)
        link.node = all_entries[entry_index]
        construct_link(link_struct, link)

    for i, entry_struct in enumerate(root.acquire("EntryList", GFFList())):
        entry = all_entries[i]
        entry.speaker = entry_struct.acquire("Speaker", "")
        construct_node(entry_struct, entry)

        for link_struct in entry_struct.acquire("RepliesList", GFFList()):
            link = DLGLink()
            link.node = all_replies[link_struct.acquire("Index", 0)]
            link.is_child = bool(link_struct.acquire("IsChild", 0))
            link.comment = link_struct.acquire("LinkComment", 0)

            entry.links.append(link)
            construct_link(link_struct, link)

    for i, reply_struct in enumerate(root.acquire("ReplyList", GFFList())):
        reply = all_replies[i]
        construct_node(reply_struct, reply)

        for link_struct in reply_struct.acquire("EntriesList", GFFList()):
            link = DLGLink()
            link.node = all_entries[link_struct.acquire("Index", 0)]
            link.is_child = bool(link_struct.acquire("IsChild", 0))
            link.comment = link_struct.acquire("LinkComment", 0)

            reply.links.append(link)
            construct_link(link_struct, link)

    return dlg


def dismantle_dlg(
        dlg: DLG,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    def dismantle_link(
            gff_struct: GFFStruct,
            link: DLGLink,
            nodes: List
    ):
        gff_struct.set_resref("Active", link.active1)
        gff_struct.set_uint32("Index", nodes.index(link.node))
        if game == Game.K2:
            gff_struct.set_resref("Active2", link.active2)
            gff_struct.set_int32("Logic", link.logic)
            gff_struct.set_uint8("Not", link.active1_not)
            gff_struct.set_uint8("Not2", link.active2_not)
            gff_struct.set_int32("Param1", link.active1_param1)
            gff_struct.set_int32("Param2", link.active1_param2)
            gff_struct.set_int32("Param3", link.active1_param3)
            gff_struct.set_int32("Param4", link.active1_param4)
            gff_struct.set_int32("Param5", link.active1_param5)
            gff_struct.set_string("ParamStrA", link.active1_param6)
            gff_struct.set_int32("Param1b", link.active2_param1)
            gff_struct.set_int32("Param2b", link.active2_param2)
            gff_struct.set_int32("Param3b", link.active2_param3)
            gff_struct.set_int32("Param4b", link.active2_param4)
            gff_struct.set_int32("Param5b", link.active2_param5)
            gff_struct.set_string("ParamStrB", link.active2_param6)

    def dismantle_node(
            gff_struct: GFFStruct,
            node: DLGNode,
            nodes: List,
            list_name: str
    ):
        gff_struct.set_locstring("Text", node.text)
        gff_struct.set_string("Listener", node.listener)
        gff_struct.set_resref("VO_ResRef", node.vo_resref)
        gff_struct.set_resref("Script", node.script1)
        gff_struct.set_uint32("Delay", node.delay)
        gff_struct.set_string("Comment", node.comment)
        gff_struct.set_resref("Sound", node.sound)
        gff_struct.set_string("Quest", node.quest)
        gff_struct.set_int32("PlotIndex", node.plot_index)
        gff_struct.set_single("PlotXPPercentage", node.plot_xp_percentage)
        gff_struct.set_uint32("WaitFlags", node.wait_flags)
        gff_struct.set_uint32("CameraAngle", node.camera_angle)
        gff_struct.set_uint8("FadeType", node.fade_type)
        gff_struct.set_uint8("SoundExists", node.sound_exists)
        gff_struct.set_uint8("Changed", node.vo_text_changed)

        anim_list = gff_struct.set_list("AnimList", GFFList())
        for anim in node.animations:
            anim_struct = anim_list.add(0)
            anim_struct.set_uint16("Animation", anim.animation_id)
            anim_struct.set_string("Participant", anim.participant)

        if node.quest_entry is not None: gff_struct.set_uint32("QuestEntry", node.quest_entry)
        if node.fade_delay is not None: gff_struct.set_single("FadeDelay", node.fade_delay)
        if node.fade_length is not None: gff_struct.set_single("FadeLength", node.fade_length)
        if node.camera_anim is not None: gff_struct.set_uint16("CameraAnimation", node.camera_anim)
        if node.camera_id is not None: gff_struct.set_int32("CameraID", node.camera_id)
        if node.camera_fov is not None: gff_struct.set_single("CamFieldOfView", node.camera_fov)
        if node.camera_height is not None: gff_struct.set_single("CamHeightOffset", node.camera_height)
        if node.camera_effect is not None: gff_struct.set_int32("CamVidEffect", node.camera_effect)
        if node.target_height is not None: gff_struct.set_single("TarHeightOffset", node.target_height)
        if node.fade_color is not None: gff_struct.set_vector3("FadeColor", node.fade_color.bgr_vector3())

        if game == Game.K2:
            gff_struct.set_int32("ActionParam1", node.script1_param1)
            gff_struct.set_int32("ActionParam1b", node.script2_param1)
            gff_struct.set_int32("ActionParam2", node.script1_param2)
            gff_struct.set_int32("ActionParam2b", node.script2_param2)
            gff_struct.set_int32("ActionParam3", node.script1_param3)
            gff_struct.set_int32("ActionParam3b", node.script2_param3)
            gff_struct.set_int32("ActionParam4", node.script1_param4)
            gff_struct.set_int32("ActionParam4b", node.script2_param4)
            gff_struct.set_int32("ActionParam5", node.script1_param5)
            gff_struct.set_int32("ActionParam5b", node.script2_param5)
            gff_struct.set_string("ActionParamStrA", node.script1_param6)
            gff_struct.set_string("ActionParamStrB", node.script2_param6)
            gff_struct.set_resref("Script2", node.script2)
            gff_struct.set_int32("AlienRaceNode", node.alien_race_node)
            gff_struct.set_int32("Emotion", node.emotion_id)
            gff_struct.set_int32("FacialAnim", node.facial_id)
            gff_struct.set_int32("NodeID", node.node_id)
            gff_struct.set_int32("NodeUnskippable", node.unskippable)
            gff_struct.set_int32("PostProcNode", node.post_proc_node)
            gff_struct.set_int32("RecordNoVOOverri", node.record_no_vo_override)
            gff_struct.set_int32("RecordVO", node.record_vo)
            gff_struct.set_int32("VOTextChanged", node.vo_text_changed)

        link_list = gff_struct.set_list(list_name, GFFList())
        for i, link in enumerate(node.links):
            link_struct = link_list.add(i)
            dismantle_link(link_struct, link, nodes)

    all_entries = dlg.all_entries()
    all_replies = dlg.all_replies()

    gff = GFF(GFFContent.DLG)

    root = gff.root
    root.set_uint32("NumWords", dlg.word_count)
    root.set_resref("EndConverAbort", dlg.on_abort)
    root.set_resref("EndConversation", dlg.on_end)
    root.set_uint8("Skippable", dlg.skippable)
    root.set_resref("AmbientTrack", dlg.ambient_track)
    root.set_uint8("AnimatedCut", dlg.animated_cut)
    root.set_resref("CameraModel", dlg.camera_model)
    root.set_uint8("ComputerType", dlg.computer_type.value)
    root.set_int32("ConversationType", dlg.conversation_type.value)
    root.set_uint8("OldHitCheck", dlg.old_hit_check)
    root.set_uint8("UnequipHItem", dlg.unequip_hands)
    root.set_uint8("UnequipItems", dlg.unequip_items)
    root.set_string("VO_ID", dlg.vo_id)
    if game == Game.K2:
        root.set_int32("AlienRaceOwner", dlg.alien_race_owner)
        root.set_int32("PostProcOwner", dlg.post_proc_owner)
        root.set_int32("RecordNoVO", dlg.record_no_vo)
        root.set_int32("NextNodeID", dlg.next_node_id)
    if use_deprecated:
        root.set_uint32("DelayEntry", dlg.delay_entry)
        root.set_uint32("DelayReply", dlg.delay_reply)

    stunt_list = root.set_list("StuntList", GFFList())
    for stunt in dlg.stunts:
        stunt_struct = stunt_list.add(0)
        stunt_struct.set_string("Participant", stunt.participant)
        stunt_struct.set_resref("StuntModel", stunt.stunt_model)

    starting_list = root.set_list("StartingList", GFFList())
    for i, starter in enumerate(dlg.starters):
        starting_struct = starting_list.add(i)
        dismantle_link(starting_struct, starter, all_entries)

    entries_list = root.set_list("EntryList", GFFList())
    for i, entry in enumerate(all_entries):
        entries_struct = entries_list.add(i)
        entries_struct.set_string("Speaker", entry.speaker)
        dismantle_node(entries_struct, entry, all_replies, "RepliesList")

    replies_list = root.set_list("ReplyList", GFFList())
    for i, reply in enumerate(all_replies):
        replies_struct = replies_list.add(i)
        dismantle_node(replies_struct, reply, all_entries, "EntriesList")

    return gff


def read_dlg(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> DLG:
    gff = read_gff(source, offset, size)
    dlg = construct_dlg(gff)
    return dlg


def write_dlg(
        dlg: DLG,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_dlg(dlg, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_dlg(
        dlg: DLG,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_dlg(dlg, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from unittest import TestCase

from pykotor.common.geometry import Vector3
from pykotor.resource.formats.bwm import BWMBinaryReader, BWM, write_bwm, read_bwm

BINARY_TEST_FILE = "../../files/test.wok"


class TestBWM(TestCase):
    def test_binary_io(self):
        wok = BWMBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(wok)

        data = bytearray()
        write_bwm(wok, data)
        wok = read_bwm(data)
        self.validate_io(wok)

    def validate_io(self, wok: BWM):
        self.assertEqual(114, len(wok.vertices()))
        self.assertEqual(195, len(wok.faces))
        self.assertTrue(wok.faces[1].v1.distance(Vector3(12.6670, 23.8963, -1.2749)) < 1e6)
        self.assertTrue(wok.faces[1].v2.distance(Vector3(12.4444, 28.6584, -1.2750)) < 1e6)
        self.assertTrue(wok.faces[1].v3.distance(Vector3(11.3294, 18.5879, -1.2750)) < 1e6)

        face2_adj = wok.adjacencies(wok.faces[2])
        self.assertIsNone(face2_adj[0])
        self.assertIs(wok.faces[29], face2_adj[1].face)
        self.assertIs(2, face2_adj[1].edge)
        self.assertIs(wok.faces[1], face2_adj[2].face)
        self.assertIs(0, face2_adj[2].edge)

        face4_adj = wok.adjacencies(wok.faces[4])
        self.assertIs(wok.faces[30], face4_adj[0].face)
        self.assertIs(2, face4_adj[0].edge)
        self.assertIs(wok.faces[35], face4_adj[1].face)
        self.assertIs(2, face4_adj[1].edge)
        self.assertIs(wok.faces[25], face4_adj[2].face)
        self.assertIs(1, face4_adj[2].edge)

        edges = wok.edges()
        self.assertEqual(73, len(edges))

        # The following tests may fail if the algorithms used to build the aabb tree or edges change. They may, however,
        # still work ingame.
        self.assertEqual([59, 66, 73], [edges.index(edge) + 1 for edge in edges if edge.final])
        self.assertEqual(389, len(wok.aabbs()))





from __future__ import annotations

import json
from typing import Optional

from pykotor.resource.formats.twoda.twoda_data import TwoDA
from pykotor.resource.type import TARGET_TYPES, SOURCE_TYPES, ResourceReader, ResourceWriter, autoclose


class TwoDAJSONReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._json = {}
        self._twoda: Optional[TwoDA] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> TwoDA:
        self._twoda = TwoDA()
        self._json = json.loads(self._reader.read_bytes(self._reader.size()).decode())

        for row in self._json['rows']:
            row_label = row["_id"]
            del row["_id"]

            for header in row:
                if header not in self._twoda.get_headers():
                    self._twoda.add_column(header)

            self._twoda.add_row(row_label, row)

        return self._twoda


class TwoDAJSONWriter(ResourceWriter):
    def __init__(
            self,
            twoda: TwoDA,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._twoda: TwoDA = twoda
        self._json = {"rows": []}

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        for row in self._twoda:
            json_row = {"_id": row.label()}
            self._json["rows"].append(json_row)
            for header in self._twoda.get_headers():
                json_row[header] = row.get_string(header)

        json_dump = json.dumps(self._json, indent=4)
        self._writer.write_bytes(json_dump.encode())



from typing import List


class PatchLogger:
    def __init__(self):
        self.notes: List[PatchLog] = []
        self.warnings: List[PatchLog] = []
        self.errors: List[PatchLog] = []

        self.patches_completed: int = 0

    def complete_patch(self) -> None:
        self.patches_completed += 1

    def add_note(self, message: str) -> None:
        self.notes.append(PatchLog(message))

    def add_warning(self, message: str) -> None:
        self.warnings.append(PatchLog(message))

    def add_error(self, message: str) -> None:
        self.errors.append(PatchLog(message))


class PatchLog:
    def __init__(self, message: str):
        self.message: str = message



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Any, Optional, Union, Tuple

from pykotor.common.language import LocalizedString
from pykotor.common.misc import ResRef

from pykotor.resource.formats.gff import GFF, GFFStruct, GFFList, GFFFieldType
from pykotor.tslpatcher.logger import PatchLogger
from pykotor.tslpatcher.memory import PatcherMemory


# TODO: 2DAMEMORY# as field path+label, store+save


class LocalizedStringDelta(LocalizedString):
    def __init__(self, stringref: Optional[FieldValue] = None):
        super().__init__(0)
        self.stringref: Optional[FieldValue] = stringref

    def apply(self, locstring: LocalizedString, memory: PatcherMemory) -> None:
        if self.stringref is not None:
            locstring.stringref = self.stringref.value(memory, GFFFieldType.UInt32)
        for language, gender, text in self:
            locstring.set(language, gender, text)


# region Value Returners
class FieldValue(ABC):
    @abstractmethod
    def value(self, memory: PatcherMemory, field_type: GFFFieldType) -> Any:
        ...

    def validate(self, value: Any, field_type: GFFFieldType) -> Any:
        if field_type == GFFFieldType.ResRef and not isinstance(value, ResRef):
            value = ResRef(str(value))
        elif field_type == GFFFieldType.String and not isinstance(value, str):
            value = str(value)
        elif field_type.return_type() == int and isinstance(value, str):
            value = int(value)
        elif field_type.return_type() == float and isinstance(value, str):
            value = float(value)
        return value


class FieldValueConstant(FieldValue):
    def __init__(self, value: Any):
        self.stored = value

    def value(self, memory: PatcherMemory, field_type: GFFFieldType) -> Any:
        return self.validate(self.stored, field_type)


class FieldValue2DAMemory(FieldValue):
    def __init__(self, token_id: int):
        self.token_id = token_id

    def value(self, memory: PatcherMemory, field_type: GFFFieldType) -> Any:
        return self.validate(memory.memory_2da[self.token_id], field_type)


class FieldValueTLKMemory(FieldValue):
    def __init__(self, token_id: int):
        self.token_id = token_id

    def value(self, memory: PatcherMemory, field_type: GFFFieldType) -> Any:
        return self.validate(memory.memory_str[self.token_id], field_type)
# endregion


# region Modify GFF
class ModifyGFF(ABC):
    @abstractmethod
    def apply(self, container: Union[GFFStruct, GFFList], memory: PatcherMemory, logger: PatchLogger) -> None:
        ...


class AddFieldGFF(ModifyGFF):
    def __init__(self, identifier: str, label: str, field_type: GFFFieldType, value: FieldValue, path: str = "", modifiers: List[ModifyGFF] = None, index_to_list_token: Optional[int] = None):
        self.identifier: str = identifier
        self.label: str = label
        self.field_type: GFFFieldType = field_type
        self.value: FieldValue = value
        self.path: Optional[str] = path
        self.index_to_list_token: Optional[int] = index_to_list_token

        self.modifiers: List[ModifyGFF] = [] if modifiers is None else modifiers

    def apply(self, container: Union[GFFStruct, GFFList], memory: PatcherMemory, logger: PatchLogger) -> None:
        container = self._navigate_containers(container, self.path)
        if container is None:
            logger.add_warning("Parent field at '{}' does not exist or is not a List or Struct. Unable to add new Field '{}'...".format(self.path, self.label))
            return

        value = self.value.value(memory, self.field_type)

        def set_locstring():
            original = LocalizedString(0)
            value.apply(original, memory)
            container.set_locstring(self.label, original)

        def set_struct():
            if isinstance(container, GFFStruct):
                return container.set_struct(self.label, value)
            elif isinstance(container, GFFList):
                return container.add(value.struct_id)

        def set_list():
            return container.set_list(self.label, value)

        func_map = {
            GFFFieldType.Int8: lambda: container.set_int8(self.label, value),
            GFFFieldType.UInt8: lambda: container.set_uint8(self.label, value),
            GFFFieldType.Int16: lambda: container.set_int16(self.label, value),
            GFFFieldType.UInt16: lambda: container.set_uint16(self.label, value),
            GFFFieldType.Int32: lambda: container.set_int32(self.label, value),
            GFFFieldType.UInt32: lambda: container.set_uint32(self.label, value),
            GFFFieldType.Int64: lambda: container.set_int64(self.label, value),
            GFFFieldType.UInt64: lambda: container.set_uint64(self.label, value),
            GFFFieldType.Single: lambda: container.set_single(self.label, value),
            GFFFieldType.Double: lambda: container.set_double(self.label, value),
            GFFFieldType.String: lambda: container.set_string(self.label, value),
            GFFFieldType.ResRef: lambda: container.set_resref(self.label, value),
            GFFFieldType.LocalizedString: lambda: set_locstring(),
            GFFFieldType.Vector3: lambda: container.set_vector3(self.label, value),
            GFFFieldType.Vector4: lambda: container.set_vector4(self.label, value),
            GFFFieldType.Struct: lambda: set_struct(),
            GFFFieldType.List: lambda: set_list(),
        }
        container = func_map[self.field_type]()

        if self.index_to_list_token is not None and isinstance(container, GFFList):
            memory.memory_2da[self.index_to_list_token] = str(len(container) - 1)

        for add_field in self.modifiers:
            add_field.apply(container, memory, logger)

    def _navigate_containers(self, container: Union[GFFStruct, GFFList], path: str) -> GFFStruct:
        hierarchy = [container for container in path.split("\\") if container != ""]

        for step in hierarchy:
            if isinstance(container, GFFStruct):
                container = container.acquire(step, None, (GFFStruct, GFFList))
            elif isinstance(container, GFFList):
                container = container.at(int(step))

        return container


class ModifyFieldGFF(ModifyGFF):
    def __init__(self, path: str, value: FieldValue):
        self.path: str = path
        self.value: FieldValue = value

    def apply(self, container: Union[GFFStruct, GFFList], memory: PatcherMemory, logger: PatchLogger) -> None:
        navigationtuple = self._navigate_containers(container, self.path)
        if navigationtuple is None:
            logger.add_warning("Unable to find a field label matching '{}', skipping...".format(self.path))
            return

        container, label, field_type = navigationtuple
        value = self.value.value(memory, field_type)

        def set_locstring():
            if container.exists(label):
                original = container.get_locstring(label)
                value.apply(original, memory)
                container.set_locstring(label, original)
            else:
                container.set_locstring(label, value)

        func_map = {
            GFFFieldType.Int8: lambda: container.set_int8(label, value),
            GFFFieldType.UInt8: lambda: container.set_uint8(label, value),
            GFFFieldType.Int16: lambda: container.set_int16(label, value),
            GFFFieldType.UInt16: lambda: container.set_uint16(label, value),
            GFFFieldType.Int32: lambda: container.set_int32(label, value),
            GFFFieldType.UInt32: lambda: container.set_uint32(label, value),
            GFFFieldType.Int64: lambda: container.set_int64(label, value),
            GFFFieldType.UInt64: lambda: container.set_uint64(label, value),
            GFFFieldType.Single: lambda: container.set_single(label, value),
            GFFFieldType.Double: lambda: container.set_double(label, value),
            GFFFieldType.String: lambda: container.set_string(label, value),
            GFFFieldType.ResRef: lambda: container.set_resref(label, value),
            GFFFieldType.LocalizedString: lambda: set_locstring(),
            GFFFieldType.Vector3: lambda: container.set_vector3(label, value),
            GFFFieldType.Vector4: lambda: container.set_vector4(label, value),
        }
        func_map[field_type]()

    def _navigate_containers(self, container: Union[GFFStruct, GFFList], path: str) -> Optional[Tuple[GFFStruct, str, GFFFieldType]]:
        hierarchy, label = path.split("\\")[:-1], path.split("\\")[-1:][0]

        for step in hierarchy:
            if isinstance(container, GFFStruct):
                container = container.acquire(step, None, (GFFStruct, GFFList))
            elif isinstance(container, GFFList):
                container = container.at(int(step))

        if container is None:
            return None

        field_type = container.what_type(label)
        return container, label, field_type
# endregion

class ModificationsGFF:
    def __init__(self, filename: str, replace_file: bool, modifiers: List[ModifyGFF] = None, destination: str = None):
        self.filename: str = filename
        self.replace_file: bool = replace_file
        self.destination: str = destination if destination is not None else "/override/"+filename
        self.modifiers: List[ModifyGFF] = modifiers if modifiers is not None else []

    def apply(self, gff: GFF, memory: PatcherMemory, logger: PatchLogger) -> None:
        for change_field in self.modifiers:
            change_field.apply(gff.root, memory, logger)



from __future__ import annotations

import io
from typing import Optional
from xml.etree import ElementTree

from pykotor.resource.formats.lip import LIP, LIPShape
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class LIPXMLReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._lip: Optional[LIP] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> LIP:
        self._lip = LIP()

        data = self._reader.read_bytes(self._reader.size()).decode()
        xml_root = ElementTree.parse(io.StringIO(data)).getroot()

        if xml_root.tag != "lip":
            raise ValueError("The XML file that was loaded was not a valid LIP.")

        self._lip.length = float(xml_root.get("duration"))

        for subelement in xml_root:
            time = float(subelement.get("time"))
            shape = LIPShape(int(subelement.get("shape")))
            self._lip.add(time, shape)

        return self._lip


class LIPXMLWriter(ResourceWriter):
    def __init__(
            self,
            lip: LIP,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._lip = lip
        self._xml_root: ElementTree.Element = ElementTree.Element("lip")

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        self._xml_root.set("duration", str(self._lip.length))

        for keyframe in self._lip:
            ElementTree.SubElement(self._xml_root,
                                   "keyframe",
                                   time=str(keyframe.time),
                                   shape=str(keyframe.shape.value))

        ElementTree.indent(self._xml_root)
        self._writer.write_bytes(ElementTree.tostring(self._xml_root))



import os

from pykotor.common.language import LocalizedString
from pykotor.common.module import Module
from pykotor.extract.installation import Installation, SearchLocation
from pykotor.resource.formats.erf import ERF, ERFType, write_erf
from pykotor.resource.formats.gff import write_gff
from pykotor.resource.formats.lyt.lyt_auto import write_lyt
from pykotor.resource.formats.rim import read_rim, RIM
from pykotor.resource.formats.tpc import TPCTextureFormat, TPC, write_tpc
from pykotor.resource.formats.vis import write_vis
from pykotor.resource.generics.are import dismantle_are
from pykotor.resource.generics.git import dismantle_git
from pykotor.resource.generics.ifo import dismantle_ifo
from pykotor.resource.generics.pth import dismantle_pth
from pykotor.resource.generics.utd import dismantle_utd
from pykotor.resource.generics.utp import dismantle_utp
from pykotor.resource.generics.uts import dismantle_uts
from pykotor.resource.type import ResourceType
from pykotor.tools import model


def clone_module(
        root: str,
        identifier: str,
        prefix: str,
        name: str,
        installation: Installation,
        *,
        copyTextures: bool = False,
        copyLightmaps: bool = False,
        keepDoors: bool = False,
        keepPlaceables: bool = False,
        keepSounds: bool = False,
        keepPathing: bool = False
) -> None:
    installation = installation
    root = root
    identifier = identifier

    oldModule = Module(root, installation)
    newModule = ERF(ERFType.MOD)

    ifo = oldModule.info().resource()
    oldIdentifier = ifo.identifier.get()
    ifo.identifier.set(identifier)
    ifo.mod_name = LocalizedString.from_english(identifier.upper())
    ifo.tag = identifier.upper()
    ifo.area_name.set(identifier)
    ifo_data = bytearray()
    write_gff(dismantle_ifo(ifo), ifo_data)
    newModule.set("module", ResourceType.IFO, ifo_data)

    are = oldModule.are().resource()
    are.name = LocalizedString.from_english(name)
    are_data = bytearray()
    write_gff(dismantle_are(are), are_data)
    newModule.set(identifier, ResourceType.ARE, are_data)

    lyt = oldModule.layout().resource()
    vis = oldModule.vis().resource()

    if keepPathing:
        pth = oldModule.pth().resource()
        pth_data = bytearray()
        write_gff(dismantle_pth(pth), pth_data)
        newModule.set(identifier, ResourceType.PTH, pth_data)

    git = oldModule.git().resource()
    git.creatures = []
    git.encounters = []
    git.stores = []
    git.triggers = []
    git.waypoints = []
    git.cameras = []

    if keepDoors:
        for i, door in enumerate(git.doors):
            oldResname = door.resref.get()
            newResname = "{}_dor{}".format(identifier, i)
            door.resref.set(newResname)
            door.tag = newResname

            utd = oldModule.door(oldResname).resource()
            data = bytearray()
            write_gff(dismantle_utd(utd), data)
            newModule.set(newResname, ResourceType.UTD, data)
    else:
        git.doors = []

    if keepPlaceables:
        for i, placeable in enumerate(git.placeables):
            oldResname = placeable.resref.get()
            newResname = "{}_plc{}".format(identifier, i)
            placeable.resref.set(newResname)
            placeable.tag = newResname

            utp = oldModule.placeable(oldResname).resource()
            data = bytearray()
            write_gff(dismantle_utp(utp), data)
            newModule.set(newResname, ResourceType.UTP, data)
    else:
        git.placeables = []

    if keepSounds:
        for i, sound in enumerate(git.sounds):
            oldResname = sound.resref.get()
            newResname = "{}_snd{}".format(identifier, i)
            sound.resref.set(newResname)
            sound.tag = newResname

            uts = oldModule.sound(oldResname).resource()
            data = bytearray()
            write_gff(dismantle_uts(uts), data)
            newModule.set(newResname, ResourceType.UTS, data)
    else:
        git.sounds = []

    git_data = bytearray()
    write_gff(dismantle_git(git), git_data)
    newModule.set(identifier, ResourceType.GIT, git_data)

    newLightmaps = {}
    newTextures = {}
    for room in lyt.rooms:
        oldModelName = room.model
        newModelName = oldModelName.lower().replace(oldIdentifier, identifier)

        room.model = newModelName
        if vis.room_exists(oldModelName):
            vis.rename_room(oldModelName, newModelName)

        mdlData = installation.resource(oldModelName, ResourceType.MDL).data
        mdxData = installation.resource(oldModelName, ResourceType.MDX).data
        wokData = installation.resource(oldModelName, ResourceType.WOK).data

        if copyTextures:
            for texture in model.list_textures(mdlData):
                if texture not in newTextures:
                    newTextureName = prefix + texture[3:]
                    newTextures[texture] = newTextureName

                    tpc = installation.texture(texture)
                    tpc = TPC() if tpc is None else tpc
                    rgba = tpc.convert(TPCTextureFormat.RGBA)

                    tga = TPC()
                    tga.set(rgba.width, rgba.height, [rgba.data], TPCTextureFormat.RGBA)

                    tga_data = bytearray()
                    write_tpc(tga, tga_data, ResourceType.TGA)
                    newModule.set(newTextureName, ResourceType.TGA, tga_data)
            mdlData = model.change_textures(mdlData, newTextures)

        if copyLightmaps:
            for lightmap in model.list_lightmaps(mdlData):
                if lightmap not in newLightmaps:
                    newLightmapName = "{}_lm_{}".format(identifier, len(newLightmaps.keys()))
                    newLightmaps[lightmap] = newLightmapName

                    tpc = installation.texture(lightmap, [SearchLocation.CHITIN, SearchLocation.OVERRIDE])
                    tpc = TPC() if tpc is None else tpc
                    rgba = tpc.convert(TPCTextureFormat.RGBA)

                    tga = TPC()
                    tga.set(rgba.width, rgba.height, [rgba.data], TPCTextureFormat.RGBA)

                    tga_data = bytearray()
                    write_tpc(tga, tga_data, ResourceType.TGA)
                    newModule.set(newLightmapName, ResourceType.TGA, tga_data)
            mdlData = model.change_lightmaps(mdlData, newLightmaps)

        mdlData = model.rename(mdlData, newModelName)
        newModule.set(newModelName, ResourceType.MDL, mdlData)
        newModule.set(newModelName, ResourceType.MDX, mdxData)
        newModule.set(newModelName, ResourceType.WOK, wokData)

    vis_data = bytearray()
    write_vis(vis, vis_data)
    newModule.set(identifier, ResourceType.VIS, vis_data)

    lyt_data = bytearray()
    write_lyt(lyt, lyt_data)
    newModule.set(identifier, ResourceType.LYT, lyt_data)

    filepath = installation.module_path() + identifier + ".mod"
    write_erf(newModule, filepath)


def rim_to_mod(filepath: str) -> None:
    """
    Creates a MOD file at the given filepath and copies the resources from the corresponding
    RIM files.

    Raises:
        ValueError: If the file was corrupted or the format could not be determined.
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.

    Args:
        filepath: The filepath of the MOD file you would like to create.
    """
    if not filepath.endswith(".mod"):
        raise ValueError("Specified file must end with the .mod extension")

    filepath_rim_s = filepath.replace(".mod", "_s.rim")
    filepath_rim = filepath.replace(".mod", ".rim")

    rim = read_rim(filepath_rim)
    rim_s = read_rim(filepath_rim_s) if os.path.exists(filepath_rim_s) else RIM()

    mod = ERF(ERFType.MOD)
    [mod.set(res.resref.get(), res.restype, res.data) for res in rim]
    [mod.set(res.resref.get(), res.restype, res.data) for res in rim_s]

    write_erf(mod, filepath, ResourceType.ERF)



import platform
from unittest import TestCase

from pykotor.resource.formats.lip import LIP, LIPShape, LIPBinaryReader, detect_lip, write_lip, LIPXMLReader, read_lip
from pykotor.resource.type import ResourceType

BINARY_TEST_FILE = "../../files/test.lip"
XML_TEST_FILE = "../../files/test.lip.xml"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_BINARY_TEST_FILE = "../../files/test_corrupted.lip"
CORRUPT_XML_TEST_FILE = "../../files/test_corrupted.lip.xml"


class TestLIP(TestCase):
    def test_binary_io(self):
        self.assertEqual(detect_lip(BINARY_TEST_FILE), ResourceType.LIP)

        lip = LIPBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(lip)

        data = bytearray()
        write_lip(lip, data, ResourceType.LIP)
        lip = LIPBinaryReader(data).load()
        self.validate_io(lip)

    def test_xml_io(self):
        self.assertEqual(detect_lip(XML_TEST_FILE), ResourceType.LIP_XML)

        lip = LIPXMLReader(XML_TEST_FILE).load()
        self.validate_io(lip)

        data = bytearray()
        write_lip(lip, data, ResourceType.LIP_XML)
        lip = LIPXMLReader(data).load()
        self.validate_io(lip)

    def validate_io(self, lip: LIP):
        self.assertAlmostEqual(lip.length, 1.50, 3)
        self.assertEqual(LIPShape.EE, lip.get(0).shape)
        self.assertEqual(LIPShape.OOH, lip.get(1).shape)
        self.assertEqual(LIPShape.TH, lip.get(2).shape)
        self.assertAlmostEqual(0.0, lip.get(0).time, 4)
        self.assertAlmostEqual(0.7777, lip.get(1).time, 4)
        self.assertAlmostEqual(1.25, lip.get(2).time, 4)

    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_lip, ".")
        else:
            self.assertRaises(IsADirectoryError, read_lip, ".")
        self.assertRaises(FileNotFoundError, read_lip, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_lip, CORRUPT_BINARY_TEST_FILE)
        self.assertRaises(ValueError, read_lip, CORRUPT_XML_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_lip, LIP(), ".", ResourceType.LIP)
        else:
            self.assertRaises(IsADirectoryError, write_lip, LIP(), ".", ResourceType.LIP)
        self.assertRaises(ValueError, write_lip, LIP(), ".", ResourceType.INVALID)



from typing import Optional

from pykotor.resource.generics.utd import UTD

from pykotor.extract.installation import Installation
from pykotor.resource.formats.twoda import TwoDA
from pykotor.resource.generics.utp import UTP
from pykotor.resource.type import ResourceType


def get_model(
        utd: UTD,
        installation: Installation,
        *,
        genericdoors: Optional[TwoDA] = None,
) -> str:
    """
    Returns the model name for the given door.

    If no value is specified for the genericdoor parameters then it will be loaded from the given installation.

    Args:
        utd: UTD object of the target door.
        installation: The relevant installation.
        genericdoors: The genericdoors.2da loaded into a TwoDA object.

    Returns:
        Returns the model name for the door.
    """
    if genericdoors is None:
        genericdoors = installation.resource("placeables", ResourceType.TwoDA)

    model = genericdoors.get_row(utd.appearance_id).get_string("modelname")

    return model



from __future__ import annotations

from enum import IntEnum
from typing import List, Optional, Tuple, Set

from pykotor.common.geometry import Vector3, Vector4, Vector2, SurfaceMaterial
from pykotor.common.misc import Color
from pykotor.resource.type import ResourceType


class MDL:
    """
    Represents a MDL/MDX file.

    Attributes:
        root: The root node of the model.
        anims: The animations stored in the model.
        name: The model name.
        fog: If fog affects the model.
        supermodel: Name of another model resource to import extra data from.
    """

    BINARY_TYPE = ResourceType.MDL

    def __init__(
            self
    ):
        self.root: MDLNode = MDLNode()
        self.anims: List[MDLAnimation] = []
        self.name: str = ""
        self.fog: bool = False
        self.supermodel: str = ""

    def get(
            self,
            node_name: str
    ) -> Optional[MDLNode]:
        pick = None

        nodes = [self.root]
        while nodes:
            node = nodes.pop()
            if node.name == node_name:
                pick = node
            else:
                nodes.extend(node.children)

        return pick

    def all_nodes(
            self
    ) -> List[MDLNode]:
        nodes = []
        scan = [self.root]
        while scan:
            node = scan.pop()
            nodes.append(node)
            scan.extend(node.children)
        return nodes

    def find_parent(
            self,
            child: MDLNode
    ) -> Optional[MDLNode]:
        all_nodes = self.all_nodes()
        parent = None
        for node in all_nodes:
            if child in node.children:
                parent = node
        return parent

    def global_position(
            self,
            node: MDLNode
    ) -> Vector3:
        position = node.position
        parent = self.find_parent(node)
        while parent is not None:
            position += parent.position
            parent = self.find_parent(parent)
        return position

    def get_by_node_id(
            self,
            node_id
    ) -> MDLNode:
        for node in self.all_nodes():
            if node.node_id == node_id:
                return node
        raise ValueError

    def all_textures(
            self
    ) -> Set[str]:
        textures = set()
        for node in self.all_nodes():
            if node.mesh and node.mesh.texture_1 != "NULL" and node.mesh.texture_1 != "":
                textures.add(node.mesh.texture_1)
        return textures

    def all_lightmaps(
            self
    ) -> Set[str]:
        lightmaps = set()
        for node in self.all_nodes():
            if node.mesh and node.mesh.texture_2 != "NULL" and node.mesh.texture_2 != "":
                lightmaps.add(node.mesh.texture_2)
        return lightmaps


# region Animation Data
class MDLAnimation:
    def __init__(
            self
    ):
        self.name: str = ""
        self.root_model: str = ""
        self.anim_length: float = 0.0
        self.transition_length: float = 0.0
        self.events: List[MDLEvent] = []
        self.root: MDLNode = MDLNode()

    def all_nodes(
            self
    ) -> List[MDLNode]:
        nodes = []
        scan = [self.root]
        while scan:
            node = scan.pop()
            nodes.append(node)
            scan.extend(node.children)
        return nodes


class MDLEvent:
    def __init__(
            self
    ):
        self.activation_time: float = 0.0
        self.name: str = ""


# endregion


# region Node Headers
class MDLNodeFlags(IntEnum):
    HEADER = 0x00000001
    LIGHT = 0x00000002
    EMITTER = 0x00000004
    CAMERA = 0x00000008
    REFERENCE = 0x00000010
    MESH = 0x00000020
    SKIN = 0x00000040
    DANGLY = 0x00000100
    AABB = 0x00000200
    SABER = 0x00000800


class MDLNode:
    """
    A node in the MDL tree that can store additional nodes or some extra data related to the model such as geometry or
    lighting.

    Attributes:
        children: List of children linked to the node.
        controllers: List of controllers linked to the node.
        name: Name of the node.
        position: The position of the node.
        orientation: The orientation of the node.
        light: Light data associated with the node.
        emitter: Emitter data associated with the node.
        mesh: Trimesh data associated with the node.
        skin: Skin data associated with the node.
        dangly: Danglymesh data associated with the node.
        aabb: Walkmesh data associated with the node
        saber: Sabermesh data associated with the node.
    """

    def __init__(
            self
    ):
        # self.parent: Optional[MDLNode] = None

        self.children: List[MDLNode] = []
        self.controllers: List[MDLController] = []
        self.name: str = ""
        self.node_id: int = -1
        self.position: Vector3 = Vector3.from_null()
        self.orientation: Vector4 = Vector4(0, 0, 0, 1)

        self.light: Optional[MDLLight] = None
        self.emitter: Optional[MDLEmitter] = None
        self.reference: Optional[MDLReference] = None
        self.mesh: Optional[MDLMesh] = None
        self.skin: Optional[MDLSkin] = None
        self.dangly: Optional[MDLDangly] = None
        self.aabb: Optional[MDLWalkmesh] = None
        self.saber: Optional[MDLSaber] = None

    def descendants(
            self
    ) -> List[MDLNode]:
        ancestors = []
        for child in self.children:
            ancestors.append(child)
            ancestors.extend(child.descendants())
        return ancestors

    def child(
            self,
            name
    ) -> MDLNode:
        for child in self.children:
            if child.name == name:
                return child
        else:
            raise KeyError


class MDLLight:
    """
    Light data that can be attached to a node.

    Attributes:
        flare_radius:
        light_priority:
        ambient_only:
        dynamic_type:
        shadow:
        flare:
        fading_light:
    """

    def __init__(
            self
    ):
        # TODO: Make enums, check if bools, docs, merge flare data into class
        self.flare_radius: float = 0.0
        self.light_priority: int = 0
        self.ambient_only: int = 0
        self.dynamic_type: int = 0
        self.shadow: int = 0
        self.flare: int = 0
        self.fading_light: int = 0
        self.flare_sizes: List = []
        self.flare_positions: List = []
        self.flare_color_shifts: List = []
        self.flare_textures: List = []


class MDLEmitter:
    """
    Emitter data that can be attached to a node.

    Attributes:
        dead_space:
        blast_radius:
        blast_length:
        branch_count:
        control_point_smoothing:
        x_grid:
        y_grid:
        spawn_type:
        update:
        render:
        blend:
        texture:
        chunk_name:
        two_sided_texture:
        loop:
        render_order:
        frame_blender:
        depth_texture:
    """

    def __init__(
            self
    ):
        # TODO: Make enums, check if bools, docs, seperate flags into booleans
        self.dead_space: float = 0.0
        self.blast_radius: float = 0.0
        self.blast_length: float = 0.0
        self.branch_count: int = 0
        self.control_point_smoothing: float = 0.0
        self.x_grid: int = 0
        self.y_grid: int = 0
        self.spawn_type: int = 0
        self.update: str = ""
        self.render: str = ""
        self.blend: str = ""
        self.texture: str = ""
        self.chunk_name: str = ""
        self.two_sided_texture: int = 0
        self.loop: int = 0
        self.render_order: int = 0
        self.frame_blender: int = 0
        self.depth_texture: str = ""
        self.flags: int = 0


class MDLReference:
    """
    Reference data that can be attached to a node.

    Attributes:
        model:
        reattachable:
    """

    def __init__(
            self
    ):
        # TODO: docs
        self.model: str = ""
        self.reattachable: bool = False


class MDLMesh:
    """
    Mesh data that can be attached to a node.
    """

    def __init__(
            self
    ):
        # TODO: look at mesh inverted counter array, rename boolean flags
        self.faces: List[MDLFace] = []
        self.diffuse: Color = Color.WHITE
        self.ambient: Color = Color.WHITE
        self.transparency_hint: int = 0
        self.texture_1: str = ""
        self.texture_2: str = ""
        self.saber_unknowns: Tuple[int, int, int, int, int, int, int, int] = (3, 0, 0, 0, 0, 0, 0, 0)
        self.animate_uv: bool = False

        self.radius: float = 0.0
        self.bb_min: Vector3 = Vector3.from_null()
        self.bb_max: Vector3 = Vector3.from_null()
        self.average: Vector3 = Vector3.from_null()
        self.area: float = 0.0

        self.uv_direction_x: float = 0.0
        self.uv_direction_y: float = 0.0
        self.uv_jitter: float = 0.0
        self.uv_jitter_speed: float = 0.0

        self.has_lightmap: bool = False
        self.rotate_texture: bool = False
        self.background_geometry: bool = False
        self.shadow: bool = False
        self.beaming: bool = False
        self.render: bool = True

        # Trimesh
        self.vertex_positions: List[Vector3] = []
        self.vertex_normals: Optional[List[Vector3]] = None
        self.vertex_uv1: Optional[List[Vector2]] = None
        self.vertex_uv2: Optional[List[Vector2]] = None

        # KotOR 2 Only
        self.dirt_enabled: bool = False
        self.dirt_texture: int = 0
        self.dirt_coordinate_space: int = 0
        self.hide_in_hologram: bool = False

    def gen_normals(
            self
    ):
        ...


class MDLSkin:
    """
    Skin data that can be attached to a node.
    """

    def __init__(
            self
    ):
        self.bone_indices: Tuple[int, int, int, int, int, int, int, int, int, int, int, int, int, int, int, int] = (
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        self.qbones: List[Vector3] = []
        self.tbones: List[Vector3] = []
        self.bonemap: List[int] = []

        self.vertex_bones: List[MDLBoneVertex] = []


class MDLDangly:
    """
    Dangly data that can be attached to a node.
    """


class MDLWalkmesh:
    """
    AABB data that can be attached to a node.
    """


class MDLSaber:
    """
    Saber data that can be attached to a node.
    """


# endregion


# region Geometry Data
class MDLBoneVertex:
    def __init__(
            self
    ):
        self.vertex_weights: Tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)
        self.vertex_indices: Tuple[float, float, float, float] = (-1.0, -1.0, -1.0, -1.0)


class MDLFace:
    def __init__(
            self
    ):
        self.v1: int = 0
        self.v2: int = 0
        self.v3: int = 0
        self.material: SurfaceMaterial = SurfaceMaterial.GRASS
        self.a1: int = 0
        self.a2: int = 0
        self.a3: int = 0
        self.coefficient: int = 0
        self.normal: Vector3 = Vector3.from_null()


# endregion


# region Controller Data
class MDLControllerType(IntEnum):
    INVALID = -1

    POSITION = 8
    ORIENTATION = 20
    SCALE = 36
    ILLUM_COLOR = 100
    ALPHA = 132

    P2P_BEZIER_2 = 132


class MDLController:
    """
    A controller is an object that gets attached to the node and influences some sort of change that is either static
    or animated.
    """

    def __init__(
            self
    ):
        self.controller_type: MDLControllerType = MDLControllerType.INVALID
        self.rows: List[MDLControllerRow] = []


class MDLControllerRow:
    def __init__(
            self,
            time,
            data
    ):
        self.time: float = time
        self.data: List[float] = data

    def __repr__(
            self
    ):
        return "MDLControllerRow({}, {})".format(self.time, self.data)

    def __str__(
            self
    ):
        return "{} {}".format(self.time, self.data).replace(',', '').replace('[', '').replace(']', '')
# endregion



from unittest import TestCase

from pykotor.common.language import Language
from pykotor.extract.talktable import TalkTable

TEST_FILE = "../../tests/files/test.tlk"


class TestTalkTable(TestCase):
    def test_size(self):
        talktable = TalkTable(TEST_FILE)
        self.assertEqual(3, talktable.size())

    def test_string(self):
        talktable = TalkTable(TEST_FILE)
        self.assertEqual("abcdef", talktable.string(0))
        self.assertEqual("ghijklmnop", talktable.string(1))
        self.assertEqual("qrstuvwxyz", talktable.string(2))
        self.assertEqual("", talktable.string(-1))
        self.assertEqual("", talktable.string(3))

    def test_voiceover(self):
        talktable = TalkTable(TEST_FILE)
        self.assertEqual("resref01", talktable.sound(0))
        self.assertEqual("resref02", talktable.sound(1))
        self.assertEqual("", talktable.sound(2))
        self.assertEqual("", talktable.sound(-1))
        self.assertEqual("", talktable.sound(3))

    def test_batch(self):
        talktable = TalkTable(TEST_FILE)
        batch = talktable.batch([2, 0, -1, 3])
        self.assertEqual(("abcdef", "resref01"), batch[0])
        self.assertEqual(("qrstuvwxyz", ""), batch[2])
        self.assertEqual(("", ""), batch[-1])
        self.assertEqual(("", ""), batch[3])

    def test_language(self):
        talktable = TalkTable(TEST_FILE)
        self.assertEqual(talktable.language(), Language.ENGLISH)



from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.twoda import TwoDA, TwoDABinaryReader, TwoDABinaryWriter, TwoDACSVWriter, TwoDACSVReader, \
    TwoDAJSONReader, TwoDAJSONWriter
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceType


def detect_2da(
        source: SOURCE_TYPES,
        offset: int = 0
) -> ResourceType:
    """
    Returns what format the TwoDA data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the TwoDA data.
        offset: Offset into the data.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.

    Returns:
        The format of the TwoDA data.
    """

    def check(
            first4
    ):
        if first4 == "2DA ":
            return ResourceType.TwoDA
        elif "{" in first4:
            return ResourceType.TwoDA_JSON
        elif "," in first4:
            return ResourceType.TwoDA_CSV
        else:
            return ResourceType.INVALID

    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                file_format = check(reader.read_string(4))
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = check(source[:4].decode('ascii', 'ignore'))
        elif isinstance(source, BinaryReader):
            file_format = check(source.read_string(4))
            source.skip(-4)
        else:
            file_format = ResourceType.INVALID
    except (FileNotFoundError, PermissionError, IsADirectoryError) as e:
        raise e
    except IOError:
        file_format = ResourceType.INVALID

    return file_format


def read_2da(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> TwoDA:
    """
    Returns an TwoDA instance from the source. The file format (TwoDA, TwoDA_CSV, TwoDA_JSON) is automatically
    determined before parsing the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
        An TwoDA instance.
    """
    file_format = detect_2da(source, offset)

    if file_format is ResourceType.INVALID:
        raise ValueError("Failed to determine the format of the GFF file.")

    if file_format == ResourceType.TwoDA:
        return TwoDABinaryReader(source, offset, size).load()
    elif file_format == ResourceType.TwoDA_CSV:
        return TwoDACSVReader(source, offset, size).load()
    elif file_format == ResourceType.TwoDA_JSON:
        return TwoDAJSONReader(source, offset, size).load()


def write_2da(
        twoda: TwoDA,
        target: TARGET_TYPES,
        file_format: ResourceType = ResourceType.TwoDA
) -> None:
    """
    Writes the TwoDA data to the target location with the specified format.

    Currently, the supported formats are: TwoDA, TwoDA_CSV and TwoDA_JSON.

    Args:
        twoda: The TwoDA file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.TwoDA:
        TwoDABinaryWriter(twoda, target).write()
    elif file_format == ResourceType.TwoDA_CSV:
        TwoDACSVWriter(twoda, target).write()
    elif file_format == ResourceType.TwoDA_JSON:
        TwoDAJSONWriter(twoda, target).write()
    else:
        raise ValueError("Unsupported format specified; use TwoDA, TwoDA_CSV or TwoDA_JSON.")


def bytes_2da(
        twoda: TwoDA,
        file_format: ResourceType = ResourceType.TwoDA
) -> bytes:
    """
    Returns the TwoDA data in the specified format (TwoDA, TwoDA_CSV or TwoDA_JSON) as a bytes object.

    This is a convience method that wraps the write_2da() method.

    Args:
        twoda: The target TwoDA object.
        file_format: The file format.

    Raises:
        ValueError: If the specified format was unsupported.

    Returns:
        The TwoDA data.
    """
    data = bytearray()
    write_2da(twoda, data, file_format)
    return data



from typing import Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.lyt import LYT, LYTAsciiWriter, LYTAsciiReader
from pykotor.resource.type import ResourceType


def read_lyt(
        source: Union[str, bytes, bytearray, BinaryReader],
        offset: int = 0,
        size: int = None
) -> LYT:
    """
    Returns an LYT instance from the source. The file format (LYT only) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
        An LYT instance.
    """
    return LYTAsciiReader(source, offset, size).load()


def write_lyt(
        lyt: LYT,
        target: Union[str, bytearray, BinaryReader],
        file_format: ResourceType = ResourceType.LYT
) -> None:
    """
    Writes the LYT data to the target location with the specified format (LYT only).

    Args:
        lyt: The LYT file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.LYT:
        LYTAsciiWriter(lyt, target).write()
    else:
        raise ValueError("Unsupported format specified; use LYT.")


def bytes_lyt(
        lyt: LYT,
        file_format: ResourceType = ResourceType.LYT
) -> bytes:
    """
    Returns the LYT data in the specified format (LYT only) as a bytes object.

    This is a convience method that wraps the write_lyt() method.

    Args:
        lyt: The target LYT.
        file_format: The file format.

    Raises:
       ValueError: If the specified format was unsupported.

    Returns:
        The LYT data.
    """
    data = bytearray()
    write_lyt(lyt, data, file_format)
    return data
